#if __cplusplus >= 201703L // >= C++17

// $Id: op_new_a_t.cc 2742 2022-09-17 21:59:20Z teshields $

// This file is part of the Libnew package that provides replacement
// implementations for core standard C library memory management
// functions, and for the standard C++ 'new()' and 'delete()' operators.
//
// Copyright (C) 2020-2022  Thomas E. Shields
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#if defined (TRACE)
#include <iostream>
#endif

#include <new>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#elif defined (USE_DMALLOC)
#include <dmalloc.h>
#endif

void* operator new (std::size_t sz, std::align_val_t al) noexcept (false)
{
  void* p;
  std::size_t align = (std::size_t) al;

#if defined (TRACE)
  std::cerr << "operator new (size_t, align_val_t) noexcept (false)"
               ": allocating " << sz << " bytes";
#endif

  if (sz == 0) sz = 1; // treat a 0-byte request as a 1-byte request

  // attempt to allocate size bytes;
  while (
#if defined (USE_BDWGC)
         GC_posix_memalign (&p, align, sz) != 0
#elif defined (USE_DMALLOC)
         (p = dmalloc_malloc(__FILE__, __LINE__,
                             sz,
                             DMALLOC_FUNC_MEMALIGN,
                             align, 0)) == nullptr
#else
         posix_memalign (&p, align, sz) != 0
#endif
        )
    {
      // allocation was unsuccessful;
      // get the current new handler function
      std::new_handler handler = std::get_new_handler ();

      if (handler == nullptr)     // if no handler is registered
         throw std::bad_alloc (); // throw an exception
      handler ();                 // hander is registered, call it
    }

  // the allocation was successful
#if defined (TRACE)
  std::cerr << " @ " << p << std::endl;
#endif
  return p; // return a pointer to the memory

}

#endif // __cplusplus >= 201703L // >= C++17
