/* $Id: codegen.cc 2841 2022-12-13 01:08:03Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2020-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <sstream>

#include "codegen.hh"
#include "ctc.hh"
#include "depGs.hh"
#include "fileMgr.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "patchlvl.hh"
#include "scan.hh"
#include "symTrans.hh"
#include "treeviz.hh"



symDeclSeq global_symDeclSeq; // initially an empty sequence
userArgSeq global_userArgSeq; // initially an empty sequence


// if "%destructor { ... } <*>" or "%printer { ... } <*>" occurs ...
bool needDestructorForStartSymbol = false;
bool needPrinterForStartSymbol = false;


void
finishYfileDefinitionSection (sortCollection& sortTable)
{
  const yoxControl& ctrl = outMgr::ctrl;

  yoxOutHrequires.putOxAttrbsUnionDecl (sortTable);

  yoxOutHrequires.skeletonText1 ();
  yoxOutHrequires.skeletonText2 ();

  if (unionDefined ())
    yoxOutHrequires.putYaccAttrbsDecl ();

  yoxOutHrequires.putOxAttrbsDecl ();

  yoxOutHrequires.skeletonText3 ();

  if (ctrl.targetParserLanguage == languageCXX &&
      ctrl.targetParserGen == bisonParserGen)
    yoxOutHprovides.putBisonCXXdefines ();

  if (ctrl.genTreeViz)
    yoxOutHprovides.putTreeVizExtDecl ();

  // 'msta' hack:
  // old versions of `msta' don't place the contents of an '%import'
  // code block at the front of the output parser file if the '%import'
  // code block is preceded by anything in the parser specification
  if (!yoxOutGimport.empty ())
    yoxOutG.append (yoxOutGimport);
  // end of 'msta' hack

  // put the Ox parser header into the Yacc output file
  yoxOutG.putParserHeader ();

  if (!ctrl.buildHeaderFile)
    {
      // append to YaccSpec if no OxHeader
      if (ctrl.targetParserGen == bisonParserGen)
        {
          if (!yoxOutHprovides.empty ())
            yoxOutG.gen ()
                .put ("%code provides\n{\n")
                .append (yoxOutHprovides)
                .gen ()
                .put ("}\n");

          if (!yoxOutH.empty ())
            yoxOutGdefs.gen ()
                .put ("%code requires\n{\n")
                .append (yoxOutH)
                .gen ()
                .put ("}\n");

          if (!yoxOutHrequires.empty ())
            yoxOutGdefs.gen ()
                .put ("%code requires\n{\n")
                .append (yoxOutHrequires)
                .gen ()
                .put ("}\n");
        }
      else // ctrl.targetParserGen != bisonParserGen
        {
          if (!yoxOutH.empty ())
            yoxOutGdefs.gen ()
                .put ("%{\n")
                .append (yoxOutH)
                .gen ()
                .put ("%}\n");

          if (!yoxOutHrequires.empty ())
            yoxOutGdefs.gen ()
                .put ("%{\n")
                .append (yoxOutHrequires)
                .gen ()
                .put ("%}\n");

          if (!yoxOutHprovides.empty ())
            yoxOutGdefs.gen ()
                .put ("%{\n")
                .append (yoxOutHprovides)
                .gen ()
                .put ("%}\n");
        }
    }
  else // finish the OxHeader
    {
      if (!yoxOutHprovides.empty ())
        yoxOutH.append (yoxOutHprovides);

      if (!yoxOutHrequires.empty ())
        yoxOutH.append (yoxOutHrequires);
    }

  yoxOutGdefs.gen ().put ("%");
  if (ctrl.targetParserGen == bisonParserGen)
    yoxOutGdefs.put ("code\n");
  yoxOutGdefs.put ("{\n");

  yoxOutGdefs.skeletonText5 ();

  if (ctrl.targetParserGen != bisonParserGen)
    yoxOutGdefs.gen ().put ("%");
  yoxOutGdefs.gen ().put ("}\n\n");

  yoxOutGdefs.gen ();

  if (ctrl.targetParserGen == bisonParserGen ||
      ctrl.targetParserGen == bisonParserGenOld)
    yoxOutGdefs.put ("%type <struct " + yoxPrefix ("OxAttrbs") + "> " +
                     yoxPrefix (augNstr) + "\n");
  else
    yoxOutGdefs.put ("%type <" + yoxPrefix ("OxAttrbs") + "> " +
                     yoxPrefix (augNstr) + "\n");

  if (needDestructorForStartSymbol)
    yoxOutGdefs.put ("%destructor { /* no-op */ } " + yoxPrefix (augNstr) +
                     "\n");

  if (needPrinterForStartSymbol)
    yoxOutGdefs.put ("%printer { /* no-op */ } " + yoxPrefix (augNstr) + "\n");

  yoxOutGdefs.put ("\n");

  yoxOutGdefs.putYaccUnionDecl ();

  yoxOutGdefs.append (yaccCODE); // dump out any accumulated Yacc code blocks
  yaccCODE.clear ();

  yoxOutGdefs.gen ().put ("\n%");
  if (ctrl.targetParserGen == bisonParserGen)
    yoxOutGdefs.put ("code\n");
  yoxOutGdefs.put ("{\n");

  yoxOutGdefs.skeletonText6 ();

  if (ctrl.targetParserGen != bisonParserGen)
    yoxOutGdefs.gen ().put ("%");
  yoxOutGdefs.gen ().put ("}\n\n");

  // copy the generated Yacc definition section into the Yacc output file
  yoxOutG.append (yoxOutGdefs);
}



void
finishYfileRulesSection (void)
{
  const yoxControl& ctrl = outMgr::ctrl;

  if (ctrl.targetParserLanguage == languageCXX &&
      ctrl.targetParserGen == bisonParserGen)
    yoxOutG.putBisonCXXdefines ();
}


// =============================== class codeGen ===============================

// public:

// 'YYSTYPE' and 'YYLTYPE' are only available in user & Ox-generated code
// for some versions of the 'bison' C++ skeletons
void
codeGen::putBisonCXXdefines (void)
{
  gen ();

  put ("#ifndef " + yaccPrefix ("STYPE", true) + "\n"
       "#define " + yaccPrefix ("STYPE", true) + " " +
       ctrl.user_namespace + "::" + ctrl.user_parser + "::semantic_type\n"
       "#endif\n"
       "#ifndef " + yaccPrefix ("LTYPE", true) + "\n"
       "#define " + yaccPrefix ("LTYPE", true) + " " +
       ctrl.user_namespace + "::" + ctrl.user_parser + "::location_type\n"
       "#endif\n");
}



void
codeGen::putOxAttrbsDecl (void)
{
  gen ();

  put ("typedef struct " + yoxPrefix ("OxAttrbs") + " {\n");

  if (yaccUnionDefined != 0 ||
      (ctrl.targetParserGen == bisonParserGen && bisonApiValueTypeUnionDefined))
    put ("union " + yoxPrefix ("YaccAttrbs") +
              " " + yoxPrefix ("YaccAttrbs") + ";\n");

  put ("struct " + yoxPrefix ("SemEl") +
       " *semEl;\n"
       "} " + yoxPrefix ("OxAttrbs") + ";\n");
}



void
codeGen::putOxAttrbsUnionDecl (sortCollection& sortTable)
{
  gen ();

  put ("#if defined (__cplusplus)\n"
       "#include <cstring>\n"
       "#include <new>\n"
       "#endif\n\n");

  if (ctrl.targetParserLanguage == languageCXX)
    put ("union " + yoxPrefix ("Attrbs") + " { int typeNum;\n");
  else
    put ("union " + yoxPrefix ("Attrbs") + " {\n");

  bool unionEmpty = true; // no attributes, or only @void attributes

  for (sort& theSort : sortTable)
    if (!theSort.typeName.empty ())
      {
        // attribute sort has a concrete type
        unionEmpty = false;

        if (ctrl.targetParserLanguage == languageCXX)
          // C++ restriction: a union member cannot be a class type
          put ("  char " +
               yoxPrefix ("Attrb") + std::to_string (theSort.typeNum) +
               "[sizeof (struct " +
               yoxPrefix ("T") + std::to_string (theSort.typeNum) +
               ")];\n");
        else
          put ("  struct " +
               yoxPrefix ("T") + std::to_string (theSort.typeNum) + " " +
               yoxPrefix ("Attrb") + std::to_string (theSort.typeNum) +";\n");
      }

  if (unionEmpty)
    put ("#ifndef __GNUC__\n"
         "  int __dummy__; /* avoid empty union if not GNU C/C++ */\n"
         "#endif\n");

  if (!unionEmpty && ctrl.targetParserLanguage == languageCXX)
    {
      // generate a constructor
      put ("  " + yoxPrefix ("Attrbs") + " () {\n"
           "    memset (this, 0, sizeof (union " + yoxPrefix ("Attrbs") + ")"
           ");\n"
           "  }\n"
           "  " + yoxPrefix ("Attrbs") + " (int typeNum) {\n"
           "    switch (typeNum) {\n");

      for (sort& theSort : sortTable)
        if (!theSort.typeName.empty ())
          // attribute sort has a concrete type
          put ("    case " + std::to_string (theSort.typeNum) + ":\n"
               "      new (static_cast<struct " +
                    yoxPrefix ("T") + std::to_string (theSort.typeNum) + " *>"
               "((void *)this)) (struct " +
                    yoxPrefix ("T") + std::to_string (theSort.typeNum) + ");\n"
               "      break;\n");

      put ("    default: /* no, or only `@void\', attributes */\n"
           "      /* no-op: already default constructed */\n"
           "      break;\n"
           "    }\n"
           "  }\n"

      // generate a destructor
                  " ~" + yoxPrefix ("Attrbs") + " () {\n"
                  "    switch (typeNum) {\n");

      for (sort& theSort : sortTable)
        if (!theSort.typeName.empty ())
          // attribute sort has a concrete type
          put ("    case " + std::to_string (theSort.typeNum) + ":\n"
               "      static_cast<struct " +
                    yoxPrefix ("T") + std::to_string (theSort.typeNum) + " *>"
               "((void *)this)->~" +
                    yoxPrefix ("T") + std::to_string (theSort.typeNum) +
                    " ();\n"
               "      break;\n");

      put ("    default: /* no, or only `@void\', attributes */\n"
           "      /* no-op */\n"
           "      break;\n"
           "    }\n"
           "  }\n");
    }

  put ("};\n");
}



void
codeGen::putYaccAttrbsDecl (void)
{
  gen ();

  if (ctrl.targetParserGen == bisonParserGen && bisonApiValueTypeUnionDefined)
    {
      put ("union " + yoxPrefix ("YaccAttrbs") + " {\n");

      for (const symDeclSeqElem& gsdp : global_symDeclSeq)
        {
          bool printed = false;

          for (const typeSeqElem& ts : gsdp.types)
            {
              const std::string& tp = ts.item;

              if (!tp.empty ())
                {
                  if (!printed)
                    {
                      printed = true;
                      put (" ");
                    }
                  put (" " + tp);
                }
            }

          printed = false;

          for (const std::string& sp : gsdp.syms)
            {
              if (!sp.empty ())
                {
                  if (!printed)
                    {
                      printed = true;
                      put (" " + addTokenPrefix (sp) + ";    /* " + sp);
                    }
                  else
                    put (" " + sp);
                }
            }

          put (" */\n");
        }

      put ("};\n");
    }
  else if (yaccUnionDefined != 0)
    {
      // there was at least one '%union' declaration
      put ("union " + yoxPrefix ("YaccAttrbs") + " {\n");

      append (unionBODY);

      gen ();

      put ("};\n");
    }
}



void
codeGen::putYaccUnionDecl (void)
{
  gen ();

  if (unionDefined ())
    {
      if (ctrl.targetParserGen == bisonParserGen &&
          bisonApiValueTypeUnsupported > 0)
        put ("/* there was at least one unsupported "
             "\"%define api.value.type\" definition */\n");

      if (yaccUnionDefined != 0)
        {
          // there was at least one '%union' declaration
          if (ctrl.unionTag.length () > 0) // "%union TAG { ... }"
            put ("%union " + ctrl.unionTag + " {\n");
          else
            put ("%union {\n");

          append (unionBODY);

          gen ();

          put ("struct /*" + yoxPrefix ("OxAttrbs") + "*/ {\n"
               "union /*" + yoxPrefix ("YaccAttrbs") + "*/ {\n");

          append (unionBODY);
          unionBODY.clear ();

          gen ();

          put ("} " + yoxPrefix ("YaccAttrbs") + ";\n"
               "struct " + yoxPrefix ("SemEl") + " *semEl;\n"
               "} " + yoxPrefix ("OxAttrbs") + ";\n"
               "}\n");
        }
    }
  else
    {
      // there was no '%union' declaration; create one
      put ("%union {\n"
           "struct {\n"
           "struct " + yoxPrefix ("SemEl") + " *semEl;\n"
           "} " + yoxPrefix ("OxAttrbs") + ";\n"
           "}\n");
    }
}



void
codeGen::putImmediateAttrbRef (attrbPair ref, bool GnotL)
{
  if ((ref.pos == -1) || (ref.num == -1))
    return;

  gen ();

  if (ctrl.genTreeViz && ref.num == NIOsymAttrbNum)
    {
      // NIO pseudo attribute reference

      if (GnotL)
        putParserImmediateAttrbHead (ref);
      else
        putLexerImmediateAttrbHead ();

      put (NIOsym);
    }
  else
    {
      sort& symSort = gSymTable[prodBuffer2[ref.pos].lexeme].symSort;
      sortAttrbSeq& sortSeq = *symSort.attrbSeq;

      if (ctrl.targetParserLanguage == languageCXX)
        put ("(*(struct " + yoxPrefix ("T") +
             std::to_string (symSort.typeNum) + " *)&(");

      if (GnotL)
        putParserImmediateAttrbHead (ref);
      else
        putLexerImmediateAttrbHead ();

      put (yoxPrefix ("Attrbs"));
      if (ctrl.targetParserLanguage == languageCXX)
        put ("))");
      else
        put ("." + yoxPrefix ("Attrb") + std::to_string (symSort.typeNum));
      put ("." + sortSeq[ref.num].id);
    }
}



void
codeGen::translateAndOutput (enum instRefType refType,
                             const attrbRefSeg& rsB,
                             const attrbPair target)
{
  // temporarily, point 'curPos' at a copy of 'rsb.origin'
  AGC::position const* save_curPos = curPos;
  AGC::position locPos = rsB.origin;

  curPos = &locPos;

  const std::string& buffer = rsB.buffer;
  long bufptr = 0, len = buffer.length ();

  // discard leading whitespace
  while (bufptr < len)
    {
      const char c = buffer[bufptr];
      if (std::isspace (c))
        {
          if (c == '\n')
            locPos.line++;
          ++bufptr;
        }
      else
        {
          break;
        }
    }

  while (bufptr < len)
    {
      const char c = buffer[bufptr];
      switch (c)
        {
        case attrbNumPairDelimChar:
          {
            attrbPair ref;

            ref.pos = std::stoi (buffer.substr (++bufptr));

            bufptr = buffer.find_first_of (attrbNumPairDelimChar, bufptr);
            assert (buffer[bufptr] == attrbNumPairDelimChar);

            ref.num = std::stoi (buffer.substr (++bufptr));

            bufptr = buffer.find_first_of (attrbNumPairDelimChar, bufptr);
            assert (buffer[bufptr] == attrbNumPairDelimChar);

            ++bufptr;
            switch (refType)
              {
              case imm:
                putImmediateAttrbRef (ref, true);
                break;
              case decor:
                putDecorateAttrbRef (ref, target);
                break;
              case trav:
                putTraversalAttrbRef (ref);
                break;
              }
          }
          break;
        case '#':
          if (!ctrl.noLineDirectives)
            {
              long end_lndir =
                  getLineDir (buffer.c_str () + bufptr, locPos.line);

              if (end_lndir > 0)
                {
                  // input file #line directive

                  wasOutName = false;
                  lastInlineno = locPos.line;

                  // backup line number extracted from the #line
                  // directive; it will get bumped correctly
                  --locPos.line;

                  // copy the matched #line directive to the output
                  end_lndir += bufptr;
                  while (bufptr <= end_lndir)
                    put (buffer[bufptr++]);
                  break;
                }
            }
#if (__APPLE__ && __clang__ && defined(__apple_build_version__) && \
     __clang_major__ > 11) ||                                      \
    (__GNUC__ > 6)
          // Apple Xcode 10 & 11 gcc, GNU 4, 5 & 6 gcc (tested)
          // issue warning: declaration does not declare anything
          __attribute__ ((fallthrough));
#endif
        default:
          lex ();
          put (c);
          ++bufptr;
          if (c == '\n')
            ++locPos.line;
        }
    } // while

  // restore 'curPos'
  curPos = save_curPos;
}



// Generate the type declaration for the attribute declaration list just read.
codeGen&
codeGen::putAttrbTypeDecl (const int typeNum,
                           const std::string& typeName,
                           const bool hasConcreteType)
{
  if (hasConcreteType)
    {
      gen ();

      put ("struct " + typeName);

      if (ctrl.targetParserLanguage == languageCXX)
        put (" { int typeNum;\n");
      else
        put (" {\n");

      append (attrDECL); // dump component declarations accumulated by the lexer
      attrDECL.clear ();

      gen ();

      if (ctrl.targetParserLanguage == languageCXX)
        put ("\n"
             " " + typeName + " ()"
             ": typeNum (" + std::to_string (typeNum) + ") {}");

      put ("\n};\n\n");
    }
  else
    // attribute sort has no concrete types => don't generate the empty struct
    attrDECL.clear (); // discard any lexer-accumulated component declarations

  return *this;
}



// write the leaf node-generation code into a Lex action
codeGen&
codeGen::putGenLeafCall (const std::string& token)
{
  const refCount& leafRefCount = installLeafRefCountInitList (token);

  const gSym& entry = gSymTable[token];

  gen ();

  put (yoxPrefix ("GenLeaf") + "(" +
       std::to_string (entry.symSort.nAttrbs) + "," +
       std::to_string (entry.symSort.typeNum) + "," +
       std::to_string (leafRefCount.start) + "," +
       std::to_string (leafRefCount.start + leafRefCount.length) + "," +
       (ctrl.thisLfileForReentrantParser ? "" : "&") +
       yaccPrefix ("lval") + ");");

  if (ctrl.genTreeViz)
    putGenTreeVizLeafCall (token);

  return *this;
}



// write the internal node-generation code into a Yacc action
codeGen&
codeGen::putGenIntNodeCall (void)
{
  const std::string& lhs_symbol = prodBuffer2[0].lexeme;
  const sort& symSort = gSymTable[lhs_symbol].symSort;

  gen ();

  put (yoxPrefix ("GenIntNode") + "(" +
       std::to_string (prodNumSeq - 1) + "," +
       std::to_string (prodBufferPointer2 - 1) + "," +
       std::to_string (symSort.nAttrbs) + "," +
       std::to_string (symSort.typeNum) + ",");
  putLHSsemStkElRef ();
  putRHSsemStkElRefList ();
  put (");\n");

  if (ctrl.genTreeViz)
    putGenTreeVizNodeCall (lhs_symbol);

  return *this;
}



// add the user-defined arguments to the 'yyyDecorate' call, if any
codeGen&
codeGen::putDecorateCall (const std::string& firstParam)
{
  gen ();

  put (yoxPrefix ("Decorate") + "(");

  if (ctrl.targetParserLanguage == languageCXX)
    // 'yyyDecorate ()' isn't a member function
    switch (ctrl.targetParserGen)
      {
      case bisonParserGen:
      case bisonParserGenOld:
        // Bison target needs "&yyparser"
        put ("&yyparser,");
        break;
      case mstaParserGen:
        // MSTA target needs "this"
        put ("this,");
        break;
      default: /* no-op */;
      }

  put (firstParam);

  if (!global_userArgSeq.empty ())
    {
      for (const userArgSeqElem& guap : global_userArgSeq)
        {
          if (!guap.types.empty ())
            put ("," + guap.userArg);
        }
    }

  put (");\n");

  return *this;
}



codeGen&
codeGen::putDecorateFwdDecl (void)
{
  putDecoratePrototype (); put (";\n");

  return *this;
}



codeGen&
codeGen::putDecoratePrototype (void)
{
  gen ();

  put ("static void " + yoxPrefix ("Decorate") + " (");

  if (ctrl.targetParserLanguage == languageCXX)
    // 'yyyDecorate ()' isn't a member function
    switch (ctrl.targetParserGen)
      {
      case bisonParserGen:
      case bisonParserGenOld:
        // Bison target needs "&yyparser"
        put ("class " + ctrl.user_namespace + "::" + ctrl.user_parser +
             "* yyyparser,");
        break;
      case mstaParserGen:
        // MSTA target needs "&yyparser"
        put ("class yyparser* yyyparser,");
        break;
      default: /* no-op */;
      }

  put ("bool pruning");

  // add user-defined 'yyparse()' arguments, if any
  for (const userArgSeqElem& guap : global_userArgSeq)
    {
      const typeSeq& types = guap.types;

      if (!types.empty ())
        {
          put (",");

          bool printed = false;
          for (const typeSeqElem& ts : types)
            {
              const std::string& tp = ts.item;
              if (!tp.empty ())
                {
                  if (printed)
                    put (" ");
                  put (tp);
                  printed = true;
                }
            }

          // the argument
          put (" " + guap.userArg);
        }
    }

  put (")");

  return *this;
}



// generate the LHS semantic stack element reference
codeGen&
codeGen::putLHSsemStkElRef (void)
{
  put ("(" + yoxPrefix ("OxAttrbs") + " *)"
       "&$<" + yoxPrefix ("OxAttrbs") + ">$");

  return *this;
}



// generate the RHS semantic stack element reference list
// for the current production
codeGen&
codeGen::putRHSsemStkElRefList (const std::string& sub_component)
{
  // 'prodBuffer2[i].yaccSemIndex' gives the value 'n' needed to
  // construct the Yacc semantic stack reference for the 'nth' symbol
  // in the RHS as '$n'; always use '$<yyyOxAttrbs>n' since the
  // attribute type being referenced is unknown, and may be unspecified

  for (int i = 1; i < prodBufferPointer2; i++)
    if (prodBuffer2[i].lexeme == errorSym)
      put (",NULL");
    else
      put (",&$<" + yoxPrefix ("OxAttrbs") + ">" +
           std::to_string (prodBuffer2[i].yaccSemIndex) + sub_component);

  return *this;
}



// generate the RHS semantic stack element reclaimation code
// for the current production
codeGen&
codeGen::putReclaimSemStkEls (void)
{
  gen ();

  put (yoxPrefix ("ReclaimSemStkEls") + "(" +
       std::to_string (prodBufferPointer2 - 1));
  putRHSsemStkElRefList ();
  put (");\n");

  return *this;
}



codeGen&
codeGen::putPatchlevelComment (void)
{
  gen ();

  put ("/* output from " PATCHLEVEL " */\n\n");

  return *this;
}



void
codeGen::putParserHeader (void)
{
  if (ctrl.stripOxConstructs)
    return;

  gen ();

  put ("%{\n\n");
  putPatchlevelComment ();
  putParserHeaderText ();
  put ("%}\n");
}



static const char* const yoxParserHeader[] = {
  "/* Ox-generated Attribute Grammar Evaluator parser specification\n",
  "   for Yacc-compatible parser generators\n",
  "\n",
};


static const char* const yoxGPL3Header[] = {
  "   Copyright (C) 1992-1994  Kurt M. Bischoff\n",
  "   Copyright (C) 2004-2022  Thomas E. Shields\n",
  "\n",
  "   This program is free software: you can redistribute it and/or modify\n",
  "   it under the terms of the GNU General Public License as published by\n",
  "   the Free Software Foundation, either version 3 of the License, or\n",
  "   (at your option) any later version.\n",
  "\n",
  "   This program is distributed in the hope that it will be useful,\n",
  "   but WITHOUT ANY WARRANTY; without even the implied warranty of\n",
  "   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n",
  "   GNU General Public License for more details.\n",
  "\n",
  "   You should have received a copy of the GNU General Public License\n",
  "   along with this program.  If not, see <http://www.gnu.org/licenses/>.  ",
                                                                         "*/\n",
  "\n",
};


static const char* const yoxParserExceptionHeader[] = {
  "/* As an exception to the standard GNU General Public License, you may\n",
  "   create a larger work that contains the Ox Attribute Grammar Evaluator\n",
  "   parser specification skeleton and distribute that work under terms of\n",
  "   your own choice, so long as that work isn't itself an\n",
  "   Attribute Grammar Evaluator generator using the skeleton or a\n",
  "   modified version thereof as an evaluator parser specification\n",
  "   skeleton.  */\n",
  "\n",
};


static const char* const yoxIdentifyHeader[] = {
  "/* Identify Ox output. */\n",
  "#define YYYOX 1\n",
};


static const char* const yoxLexerHeader[] = {
  "/* Ox-generated Attribute Grammar Evaluator lexer specification\n",
  "   for Lex-compatible lexer generators\n",
  "\n",
};


static const char* const yoxLexerExceptionHeader[] = {
  "/* As an exception to the standard GNU General Public License, you may\n",
  "   create a larger work that contains the Ox Attribute Grammar Evaluator\n",
  "   lexer specification skeleton and distribute that work under terms of\n",
  "   your own choice, so long as that work isn't itself an\n",
  "   Attribute Grammar Evaluator generator using the skeleton or a\n",
  "   modified version thereof as an evaluator lexer specification\n",
  "   skeleton.  */\n",
  "\n",
};



codeGen&
codeGen::putParserHeaderText (void)
{
  putStringArray (yoxParserHeader, sizeof (yoxParserHeader));
  putStringArray (yoxGPL3Header, sizeof (yoxGPL3Header));
  putStringArray (yoxParserExceptionHeader, sizeof (yoxParserExceptionHeader));
  putStringArray (yoxIdentifyHeader, sizeof (yoxIdentifyHeader));

  if (ctrl.genTreeViz)
    putTreeVizId ();

  if (ctrl.targetParserLanguage == languageC)
    put ("\n"
         "#if defined (lint)\n"
         "#include <note.h>\n"
         "#endif\n"
         "\n");

  return *this;
}



void
codeGen::putLexerHeader (enum whichYoxLexer scanner)
{
  if (scanner == LEX_SPEC_LEXER)
    {
      gen ();

      put ("%{\n\n");
    }

  putPatchlevelComment ();

  putStringArray (yoxLexerHeader, sizeof (yoxLexerHeader));
  putStringArray (yoxGPL3Header, sizeof (yoxGPL3Header));
  putStringArray (yoxLexerExceptionHeader, sizeof (yoxLexerExceptionHeader));
  putStringArray (yoxIdentifyHeader, sizeof (yoxIdentifyHeader));

  if (ctrl.genTreeViz)
    putTreeVizId ();

  if (ctrl.targetLexerLanguage == languageC)
    put ("\n"
         "#if defined (lint)\n"
         "#include <note.h>\n"
         "#endif\n"
         "\n");

  if (scanner == LEX_SPEC_LEXER)
    put ("#if defined (__cplusplus)\n"
         "#include <cstdbool>\n"
         "#else\n"
         "#include <stdbool.h>\n"
         "#endif\n"
         "%}\n");
}

// protected:

void
codeGen::putDecorateAttrbRef (attrbPair ref, attrbPair target)
{
  if ((ref.pos == -1) || (ref.num == -1))
    return;

  gen ();

  if (ctrl.genTreeViz && ref.num == NIOsymAttrbNum)
    {
      // NIO pseudo attribute reference

      put ("RSNode->");
      if (ref.pos > 0)
        put ("cL[" + std::to_string (ref.pos - 1) + "]->");
      put (NIOsym);
    }
  else
    {
      const sort& symSort = gSymTable.at (prodBuffer2[ref.pos].lexeme).symSort;
      sortAttrbSeq& sortSeq = *symSort.attrbSeq;

      if (ctrl.targetParserLanguage == languageCXX)
        put ("(*(struct " + yoxPrefix ("T") +
             std::to_string (symSort.typeNum) + " *)&(");

      if ((ref.pos == target.pos) && (ref.num == target.num))
        put ("RSTopNode->");
      else
        {
          put ("RSNode->");
          if (ref.pos > 0)
            put ("cL["+ std::to_string (ref.pos - 1) + "]->");
        }
      put (yoxPrefix ("Attrbs"));

      if (ctrl.targetParserLanguage == languageCXX)
        put ("))");
      else
        put ("." + yoxPrefix ("Attrb") + std::to_string (symSort.typeNum));
      put ("."+ sortSeq[ref.num].id);
    }
}



void
codeGen::putTraversalAttrbRef (attrbPair ref)
{
  if ((ref.pos == -1) || (ref.num == -1))
    return;

  gen ();

  if (ctrl.genTreeViz && ref.num == NIOsymAttrbNum)
    {
      // NIO pseudo attribute reference

      put ("travNode->");
      if (ref.pos > 0)
        put ("cL[" + std::to_string (ref.pos - 1) + "]->");
      put (NIOsym);
    }
  else
    {
      const sort& symSort = gSymTable[prodBuffer2[ref.pos].lexeme].symSort;
      sortAttrbSeq& sortSeq = *symSort.attrbSeq;

      if (ctrl.targetParserLanguage == languageCXX)
        put ("(*(struct " + yoxPrefix ("T") +
             std::to_string (symSort.typeNum) + " *)&(");

      put ("travNode->");
      if (ref.pos > 0)
        put ("cL[" + std::to_string (ref.pos - 1) + "]->");
      put (yoxPrefix ("Attrbs"));

      if (ctrl.targetParserLanguage == languageCXX)
        put ("))");
      else
        put ("." + yoxPrefix ("Attrb") + std::to_string (symSort.typeNum));
      put ("." + sortSeq[ref.num].id);
    }
}

// private:

// generate the correct C/C++ code for a parser stack node reference:
//      '$<yyyOxAttrbs>$.semEL->node->'
void
codeGen::putParserImmediateAttrbHead (attrbPair ref)
{
  put ("$<" + yoxPrefix ("OxAttrbs") + ">$.semEl->node->");

  if (ref.pos > 0)
    put ("cL[" + std::to_string (ref.pos - 1) + "]->");
}



// generate the correct C/C++ code for an `yylval' node reference:
//      'yylval(.|->)yyyOxAttbs.semEl->node->'
void
codeGen::putLexerImmediateAttrbHead (void)
{
  put (yaccPrefix ("lval"));

  if (ctrl.thisLfileForReentrantParser)
    put ("->");
  else
    put (".");

  if (ctrl.targetParserGen == bisonParserGen && bisonApiValueTypeUnionDefined)
    put (addTokenPrefix (yoxPrefix (augNstr)));
  else
    put (yoxPrefix ("OxAttrbs"));

  put (".semEl->node->");
}

// ================================ end codeGen ================================
