/* $Id: ctc.cc 2840 2022-12-13 01:07:00Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2007-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include "codegen.hh"
#include "ctc.hh"
#include "depGs.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "treeviz.hh"


// =============================== class codeGen ===============================

// public:

void
codeGen::putStringArray (const char* const* arr, size_t sizeofArr)
{
  gen ();

  for (int i = 0; i < (int)(sizeofArr / sizeof (char*)); i++)
    {
      bool cond = true;
      char* arrp = (char*)&arr[i][0];

      if ((*arrp == '@') && (*(arrp + 1) == '@'))
        {
          cond = false;
          arrp += 2;

          while (*arrp != '\0' && *arrp != '@')
            {
              if (((*arrp == 'u') && ctrl.showPTsizeStats) ||
                  ((*arrp == 'c') && !ctrl.suppressCycleReports) ||
                  ((*arrp == 'd') && !ctrl.suppressPruning) ||
                  ((*arrp == 'C') && !ctrl.suppressCycleReports &&
                   !ctrl.suppressPruning))
                cond = true;
              arrp++;
            }

          if (*arrp == '\0')
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "invalid string array entry: [" +
                             std::string (arr[i]) + "]\n");
            }

          arrp++;
        }

      if (cond)
        {
          if ((*arrp == '%') && (*(arrp + 1) == '@'))
            {
              arrp++;
              for (; *++arrp != '\0';)
                {
                  if (*arrp != '%')
                    put (*arrp);
                  else if (*(arrp + 1) == '%')
                    put (*++arrp); /* "%%" -> "%" */
                  else if (*(arrp + 1) == 'o')
                    {
                      arrp++;
                      put (yoxPrefix (""));
                    }
                  else if (*(arrp + 1) == 'y')
                    {
                      arrp++;
                      put (yaccPrefix (""));
                    }
                  else if (*(arrp + 1) == 'Y')
                    {
                      arrp++;
                      put (yaccPrefix ("", true));
                    }
                  else
                    {
                      const AGC::location loc; // dummy location
                      yox_fatal (loc,
                                 "invalid string array entry: [" +
                                     std::string (arr[i]) + "]\n");
                    }
                }
            }
          else
            put (arrp);
        }
    }
}



// *********************************************************************

static const char* const skeleton1[] = {
  "\n",
  "#if defined (__cplusplus)\n",
  "#include <cstdbool>\n",
  "#include <cstdlib>\n",
  "#else\n",
  "#include <stdbool.h>\n",
  "#include <stdlib.h>\n",
  "#endif\n",
  "\n",
  "%@typedef unsigned short %oWAT;\n", /* Which Attribute Type */
  "%@typedef unsigned short %oRCT;\n", /* Reference Count Type */
  "%@typedef unsigned short %oPNT;\n", /* Production Number Type */
  "%@typedef unsigned short %oWST;\n", /* Which Symbol Type */
  "%@ /* funny type; as wide as the widest of %oWAT, %oWST, %oRCT */\n",
  "%@typedef unsigned short %oFT;\n",
  "\n",
  "\n",
}; // skeleton1[]

void
codeGen::skeletonText1 (void)
{
  putStringArray (skeleton1, sizeof (skeleton1));

  if (ctrl.genTreeViz)
    putTreeVizGraphvizIncl ();
}

// *********************************************************************

static const char* const skeleton2[] = {
  "\n",
  "%@struct %oSolvedSAlistCell {%oWAT attrbNum;\n",
  "%@                            struct %oSolvedSAlistCell *next;\n",
  "                           };\n",
  "%@typedef struct %oSolvedSAlistCell %oSSALC;\n",
  "\n",
  "%@struct %oSemEl {struct %oGenNode *node;\n",
  "%@                 struct %oSolvedSAlistCell *solvedSAlist;\n",
  "                };\n",
  "%@typedef struct %oSemEl %oSIT;\n",
  "\n",
}; // skeleton2[]

void
codeGen::skeletonText2 (void)
{
  putStringArray (skeleton2, sizeof (skeleton2));
}

// *********************************************************************

static const char* const skeleton3a[] = {
  "\n",
  "%@typedef union{struct %oSemEl *stackref; struct %oGenNode *noderef;} "
      "%oParent;\n",
  "%@struct %oGenNode {%oParent parent;\n",
  "%@                   bool parentIsStack; /* =true =>  `(%oSIT *) parent\' "
      "is referenced on the Yacc semantic stack  */\n",
  "%@                                       /* =false => `(%oGNT *) parent\' "
      "is this node's parent node  */\n",
}; // skeleton3a[]

static const char* const skeleton3b[] = {
  "                   int cLlen; /* length of child list */\n",
  "%@                   struct %oGenNode **cL; /* child list */\n",
  "                   int refCountListLen; /* length of RCL */\n",
  "%@                   %oRCT *refCountList;\n",
  "%@                   %oPNT prodNum;\n",
  "%@                   %oWST whichSym; /* which child of parent? */\n",
  "%@                   union %oAttrbs %oAttrbs; /* the Ox semantic attributes",
                                                                        " */\n",
}; // skeleton3b[]

static const char* const skeleton3c[] = {
  "                  };\n",
  "%@typedef struct %oGenNode %oGNT;\n",
}; // skeleton3c[]

void
codeGen::skeletonText3 (void)
{
  putStringArray (skeleton3a, sizeof (skeleton3a));

  if (ctrl.genTreeViz)
    {
      put ("                   Agnode_t *"); put (NIOsym); put (";\n");
    }

  putStringArray (skeleton3b, sizeof (skeleton3b));

  if (ctrl.targetParserLanguage == languageCXX)
    {
      put ("                   ");
      put (yoxPrefix ("GenNode") + " (int typeNum) {\n");
      put ("                   ");
      put ("  new (&" + yoxPrefix ("Attrbs") + ")"
                   " (union " + yoxPrefix ("Attrbs") + ") (typeNum);\n");
      put ("                   }\n");
    }

  putStringArray (skeleton3c, sizeof (skeleton3c));
}

// *********************************************************************

void
codeGen::skeletonText4 (void)
{
  gen ();
  put ("extern void " + yoxPrefix ("GenLeaf") +
       " (int, int, long, long, " + yaccPrefix ("STYPE", true) + " *);\n"
       "extern void " + yoxPrefix ("abort") + " (void);\n");

  if (ctrl.genTreeViz)
    putGenTreeVizLeafExtDecl ();

  put ("extern bool " + yoxPrefix ("Yok") + ";\n");
}

// *********************************************************************

static const char* const skeleton5[] = {
  "%@struct %oRSitem {%oGNT *node;\n",
  "%@                  %oWST whichSym;\n",
  "%@                  %oWAT whichAttrb;\n",
  "                 };\n",
  "\n",
  "%@static struct %oRSitem *%oRS;\n",
  "%@static long %oRSTop = 0L;\n",
  "%@#define %oRSBottom 1L\n",
  "%@static long %oAfterRS = 1L;\n",
  "\n",
  "#include <limits.h>\n",
  "%@#define %oR USHRT_MAX\n",
}; // skeleton5[]

void
codeGen::skeletonText5 (void)
{
  putStringArray (skeleton5, sizeof (skeleton5));
}

// *********************************************************************

static const char* const skeleton6[] = {
  "%@void %oabort (void);\n",
  "\n",
  "%@static void %oExecuteRRsection (%oGNT *);\n",
  "%@static void %oDoTraversals (%oGNT *);\n",
  "%@static void %oGenIntNode (long, int, int, int, %oOxAttrbs *, ...);\n",
  "%@static void %oReclaimSemStkEls (int, ...);\n",
  "%@static void %oAdjustINRC (long, int, long, long, %oOxAttrbs *, ...);\n",
  "@@d@%@static void %oPrune (long, %oOxAttrbs *);\n",
  "@@c@%@static void %oCheckUnsolvedInstTrav (%oGNT *, long *, long *);\n",
  "@@C@%@static void %oUnsolvedInstSearchTrav (%oGNT *);\n",
  "@@c@%@static void %oUnsolvedInstSearchTravAux (%oGNT *);\n",
}; // skeleton6[]

void
codeGen::skeletonText6 (void)
{
  put ("#include <stdio.h>\n"
       "#include <stdarg.h>\n"
       "\n"
       "bool " + yoxPrefix ("Yok") + " = ");
  if (ctrl.neutralizeSemActions)
    put ("false;\n");
  else
    put ("true;\n");

  putStringArray (skeleton6, sizeof (skeleton6));

  putDecorateFwdDecl ();

  if (ctrl.genTreeViz)
    putGenTreeVizNodeFwdDecl ();
}

// *********************************************************************

static const char* const skeleton7[] = {
  "/* The maximum size of the Ready Set. */\n",
  "%@static long %oRSmaxSize =        0L;\n",
  "/* The maximum size of the postdecoration-traversal stack; */\n",
  "/* needed for a depth-first traversal of the parse tree. */\n",
  "%@static long %oTravStackMaxSize = 1000L;\n",
  "\n",
  "%@#define %oLambdaSSAL (%oSSALC *) NULL\n",
  "%@static %oSSALC *%oSSALCfreeList = %oLambdaSSAL;\n",
  "\n",
}; // skeleton7[]

void
codeGen::skeletonText7 (void)
{
  putStringArray (skeleton7, sizeof (skeleton7));
}

// *********************************************************************

static const char* const skeleton8[] = {
  "\n",
  "/* GCC (and some other compilers) define '__attribute__', but */\n",
  "/* just in case '__attribute__' isn't defined, make a dummy.  */\n",
  "#if !(defined (__GNUC__) || defined (__attribute__))\n",
  "#define __attribute__(p) /* nothing */\n",
  "#endif\n",
  "\n",
  "%@static __attribute__((__noreturn__)) void %ofatal (const char *fmt, "
      "...)\n",
  "  {va_list ap;\n",
  "   va_start (ap, fmt);\n",
  "   (void) vfprintf (stderr, fmt, ap);\n",
  "   va_end (ap);\n",
  "   exit (EXIT_FAILURE);\n",
  "#if defined (S_SPLINT_S)\n",
  "   /*@notreached@*/ ;\n",
  "#elif defined (lint)\n",
  "   NOTE(NOTREACHED)\n",
  "#elif defined (__GNUC__)\n",
  "   __builtin_unreachable ();\n",
  "#endif\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oSignalEnts (%oGNT *node, long startP, long stopP)\n",
  "%@  {%oGNT *dumNode;\n",
#if defined(DEBCTC)
  "%@   (void) fprintf (stderr, \"in %oSignalEnts: node: %lx \", (unsigned "
      "long) node);\n",
  "   (void) fprintf (stderr, \"ent list bounds: %ld %ld;  \\n\",\n",
  "                           startP, stopP\n",
  "                  );\n",
#endif
  "\n",
  "   while (startP < stopP)\n",
  "%@     {if (%oEntL[startP]==0) dumNode = node;\n",
  "%@      else dumNode = (node->cL)[%oEntL[startP]-1];\n",
#if defined(DEBCTC)
  "      (void) fprintf (stderr,\"%o %lx attrb %d         /****/   to "
      "%d\\n\",\n",
  "                             \"decrementing ref count of node \",\n",
  "%@                             (unsigned long) dumNode, %oEntL[startP "
      "+1],\n",
  "%@                             (dumNode->refCountList)[%oEntL[startP +1]] - "
      "1\n",
  "                     );\n",
#endif
  "%@      if ((--((dumNode->refCountList)[%oEntL[startP +1]]))==0)\n",
  "%@         {if (++%oRSTop==%oAfterRS)\n",
  "%@             {/* reallocate '%oRS' data structure */\n",
  "#if defined (__cplusplus)\n",
  "%@              long oldRSmaxSize = %oRSmaxSize;\n",
  "%@              %oRSmaxSize += 500L;\n",
  "              try {\n",
  "%@                 struct %oRSitem *newRS =\n",
  "%@                    new struct %oRSitem [%oRSmaxSize +1];\n",
  "                 for (int i = 0; i < oldRSmaxSize; i++)\n",
  "%@                    newRS [i] = %oRS [i];\n",
  "%@                 delete[] %oRS; %oRS = newRS;\n",
  "              } catch (std::bad_alloc) {\n",
  "%@                 %ofatal (\"error in Ox Ready Set space "
      "allocation\\n\");\n",
  "              }\n",
  "#else /* !defined (__cplusplus) */\n",
  "%@              %oRSmaxSize += 500L;\n",
  "%@              if ((%oRS =\n",
  "%@                   (struct %oRSitem *) realloc (%oRS,\n",
  "%@                                                 (size_t) ((size_t) "
      "(%oRSmaxSize +1) * sizeof (struct %oRSitem)))\n",
  "%@                  )==(struct %oRSitem *) NULL\n",
  "                 )\n",
  "%@                 %ofatal (\"realloc error in Ox Ready Set space "
      "allocation\\n\");\n",
  "#endif /* !defined (__cplusplus) */\n",
  "%@              %oAfterRS = %oRSmaxSize +1;\n",
  "             }\n",
  "%@          %oRS[%oRSTop].node = dumNode;\n",
  "%@          %oRS[%oRSTop].whichSym = %oEntL[startP];\n",
  "%@          %oRS[%oRSTop].whichAttrb = %oEntL[startP +1];\n",
#if defined(DEBCTC)
  "          (void) fprintf (stderr,\"readying:  node: %lx  whichSym: %d  "
      "whichAttrb: %d\\n\",\n",
  "%@                                 (unsigned long) dumNode, %oEntL[startP], "
      "%oEntL[startP +1]\n",
  "                         );\n",
#endif
  "         }\n",
  "      startP += 2;\n",
  "     }\n",
  "  }\n",
}; // skeleton8[]

void
codeGen::skeletonText8 (void)
{
  putStringArray (skeleton8, sizeof (skeleton8));
}

// *********************************************************************

static const char* const skeleton9a[] = {
  "\n",
  "\n",
  "\n",
  "%@static void %oGenIntNode (long ProdNum, int RHSlength, int Nattrbs,\n",
  "#if defined (S_SPLINT_S)\n",
  "                           /*@unused@*/\n",
  "#endif\n",
  "                           __attribute__((unused))\n",
  "%@                           int typeNum, %oOxAttrbs *val_OxAttrbs, ...)\n",
  "  {\n",
  "#if defined (lint)\n",
  "   NOTE(ARGUNUSED(typeNum))\n",
  "#endif\n",
  "%@   %oSIT **LHSsemEl = &val_OxAttrbs->semEl;\n",
  "\n",
  "@@u@%@   %oIntInstanceCount += Nattrbs; %oIntNodeCount++;\n",
  "#if defined (__cplusplus)\n",
  "   try {\n",
  "%@      (*LHSsemEl) = new %oSIT;\n",
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in Ox/Yacc semantic stack space ",
                                                          "allocation\\n\");\n",
  "   }\n",
  "   try {\n",
}; // skeleton9a[]

static const char* const skeleton9b[] = {
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in Ox Semantic Node space allocation\\n\");\n",
  "   }\n",
  "#else /* !defined (__cplusplus) */\n",
  "%@   (*LHSsemEl) = (%oSIT *) malloc ((size_t) sizeof (%oSIT));\n",
  "%@   if (*LHSsemEl==(%oSIT *) NULL)\n",
  "%@      %ofatal (\"malloc error in Ox/Yacc semantic stack space "
      "allocation\\n\");\n",
  "%@   (*LHSsemEl)->node = (%oGNT *) malloc ((size_t) sizeof (%oGNT));\n",
  "%@   if ((*LHSsemEl)->node==(%oGNT *) NULL)\n",
  "%@      %ofatal (\"malloc error in Ox Semantic Node space "
      "allocation\\n\");\n",
  "#endif /* !defined (__cplusplus) */\n",
  "%@   (*LHSsemEl)->solvedSAlist = %oLambdaSSAL;\n",
  "   (*LHSsemEl)->node->parent.stackref = (*LHSsemEl);\n",
  "   (*LHSsemEl)->node->parentIsStack = true;\n",
  "   (*LHSsemEl)->node->cLlen  = RHSlength;\n",
  "   if (RHSlength>0)\n",
  "#if defined (__cplusplus)\n",
  "      try {\n",
  "%@         (*LHSsemEl)->node->cL = new %oGNT * [RHSlength];\n",
  "      } catch (std::bad_alloc) {\n",
  "%@         %ofatal (\"error in Ox Child List space allocation\\n\");\n",
  "      }\n",
  "#else /* !defined (__cplusplus) */\n",
  "     {(*LHSsemEl)->node->cL =\n",
  "%@            (%oGNT **) calloc ((size_t)RHSlength, (size_t)sizeof (%oGNT "
      "*));\n",
  "%@      if ((*LHSsemEl)->node->cL==(%oGNT **) NULL)\n",
  "%@         %ofatal (\"calloc error in Ox Child List space "
      "allocation\\n\");\n",
  "     }\n",
  "#endif /* !defined (__cplusplus) */\n",
  "   else\n",
  "%@      (*LHSsemEl)->node->cL = (%oGNT **) NULL;\n",
  "   (*LHSsemEl)->node->refCountListLen = Nattrbs;\n",
  "   if (Nattrbs>0)\n",
  "#if defined (__cplusplus)\n",
  "      try {\n",
  "%@         (*LHSsemEl)->node->refCountList = new %oRCT [Nattrbs];\n",
  "         for (int i = 0; i < Nattrbs; i++)\n",
  "            (*LHSsemEl)->node->refCountList [i] = 0;\n",
  "      } catch (std::bad_alloc) {\n",
  "%@         %ofatal (\"error in Ox Reference Count List space "
      "allocation\\n\");\n",
  "      }\n",
  "#else /* !defined (__cplusplus) */\n",
  "     {(*LHSsemEl)->node->refCountList =\n",
  "%@            (%oRCT *) calloc ((size_t) Nattrbs, (size_t) sizeof "
      "(%oRCT));\n",
  "%@      if ((*LHSsemEl)->node->refCountList==(%oRCT *) NULL)\n",
  "%@         %ofatal (\"calloc error in Ox Reference Count List space "
      "allocation\\n\");\n",
  "     }\n",
  "#endif /* !defined (__cplusplus) */\n",
  "   else\n",
  "%@      (*LHSsemEl)->node->refCountList = (%oRCT *) NULL;\n",
  "%@   (*LHSsemEl)->node->prodNum = (%oPNT)ProdNum;\n",
  "   if (RHSlength>0)\n",
  "     {va_list ap;\n",
  "      va_start (ap, val_OxAttrbs);\n",
  "%@      for (%oWST i = 1; i <= RHSlength; i++)\n",
  "%@        {%oOxAttrbs *RHSattrbs = va_arg (ap, %oOxAttrbs *);\n",
  "%@         if (RHSattrbs != (%oOxAttrbs *) NULL)\n",
  "%@           {%oSIT *RHSsemEl = RHSattrbs->semEl;\n",
  "%@            %oGNT *gnpDum = (*LHSsemEl)->node->cL[i-1] = "
      "RHSsemEl->node;\n",
  "            gnpDum->whichSym = i;\n",
  "            gnpDum->parent.noderef = (*LHSsemEl)->node;\n",
  "            gnpDum->parentIsStack = false;\n",
  "           }\n",
  "        }\n",
  "      va_end (ap);\n",
  "     }\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oReclaimSemStkEls (int RHSlength, ...)\n",
  "  {if (RHSlength>0)\n",
  "     {va_list ap;\n",
  "      va_start (ap, RHSlength);\n",
  "%@      for (%oWST i = 1; i <= RHSlength; i++)\n",
  "%@        {%oOxAttrbs *RHSattrbs = va_arg (ap, %oOxAttrbs *);\n",
  "%@         if (RHSattrbs != (%oOxAttrbs *) NULL)\n",
  "%@           {%oSIT *RHSsemEl = RHSattrbs->semEl;\n",
  "%@            RHSsemEl->node = (%oGNT *) NULL;\n",
  "#if defined (__cplusplus)\n",
  "            delete RHSsemEl;\n",
  "#else\n",
  "            free (RHSsemEl);\n",
  "#endif\n",
  "           }\n",
  "        }\n",
  "      va_end (ap);\n",
  "     }\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oAdjustINRC (long ProdNum, int RHSlength, long startP, long "
      "stopP, %oOxAttrbs *val_OxAttrbs, ...)\n",
  "%@  {%oSIT *LHSsemEl = val_OxAttrbs->semEl;\n",
  "%@   %oSSALC *SSALptr,*SSALptrHead,**cPtrPtr;\n",
  "%@   %oGNT *gnpDum;\n",
  "   long nextP;\n",
  "\n",
#if defined(DEBCTC)
  "   (void) fprintf (stderr, \"initializing ref counts IN %ld %ld \\n\", "
      "startP, stopP);\n",
#endif
  "   nextP = startP;\n",
  "   while (nextP < stopP)\n",
  "%@     {if (%oRCIL[nextP]==%oR)\n",
  "%@         {(LHSsemEl->node->refCountList)[%oRCIL[nextP +1]] = "
      "%oRCIL[nextP+2];\n",
#if defined(DEBCTC)
  "          (void) fprintf (stderr,\"      %lx %d %d\\n\",\n",
  "                                 (unsigned long) LHSsemEl->node,\n",
  "%@                                 %oRCIL[nextP +1],\n",
  "%@                                 %oRCIL[nextP +2]\n",
  "                         );\n",
#endif
  "         }\n",
  "      else\n",
  "%@         "
      "{(((LHSsemEl->node->cL)[%oRCIL[nextP]])->refCountList)[%oRCIL[nextP +1]]"
      " = %oRCIL[nextP+2];\n",
#if defined(DEBCTC)
  "          (void) fprintf (stderr, \"      %lx %d %d\\n\",\n",
  "%@                                  (unsigned long) "
      "(LHSsemEl->node->cL)[%oRCIL[nextP]],\n",
  "%@                                  %oRCIL[nextP +1],\n",
  "%@                                  %oRCIL[nextP +2]\n",
  "                         );\n",
#endif
  "         }\n",
  "      nextP += 3;\n",
  "     }\n",
  "   if (RHSlength>0)\n",
  "%@     {long *pL = %oIIEL + %oIIIEL[ProdNum];\n",
  "      long iTemp;\n",
  "      va_list ap;\n",
  "      va_start (ap, val_OxAttrbs);\n",
  "%@      for (%oWST i = 1; i <= RHSlength; i++)\n",
  "%@        {%oOxAttrbs *RHSattrbs = va_arg (ap, %oOxAttrbs *);\n",
  "%@         if (RHSattrbs != (%oOxAttrbs *) NULL)\n",
  "%@           {%oSIT *RHSsemEl = RHSattrbs->semEl;\n",
  "            pL++;\n",
  "            SSALptrHead = SSALptr = *(cPtrPtr = "
      "&(RHSsemEl->solvedSAlist));\n",
  "%@            if (SSALptr != %oLambdaSSAL)\n",
  "%@              {*cPtrPtr = %oLambdaSSAL;\n",
  "               do\n",
  "                 {iTemp = (*pL+SSALptr->attrbNum);\n",
#if defined(DEBCTC)
  "                  (void) fprintf (stderr,\"%o  %d %d\",\n",
  "                                         \"\\nsignalling ents of SSA: "
      "\\n\",\n",
  "                                         i,\n",
  "                                         SSALptr->attrbNum\n",
  "                                 );\n",
#endif
  "%@                  %oSignalEnts (LHSsemEl->node,\n",
  "%@                                 %oIEL[iTemp],\n",
  "%@                                 %oIEL[iTemp +1]\n",
  "                                );\n",
  "                  SSALptr = *(cPtrPtr = &(SSALptr->next));\n",
  "                 }\n",
  "%@               while (SSALptr != %oLambdaSSAL);\n",
  "%@               *cPtrPtr = %oSSALCfreeList;\n",
  "%@               %oSSALCfreeList = SSALptrHead;\n",
  "              }\n",
  "           }\n",
  "        }\n",
  "      va_end (ap);\n",
  "     }\n",
  "   nextP = startP + 2;\n",
  "   while (nextP < stopP)\n",
  "%@     {if (%oRCIL[nextP]==0)\n",
  "%@         {if (%oRCIL[nextP-2]==%oR)\n",
  "%@             {%oSSALC **pL = &(LHSsemEl->solvedSAlist);\n",
  "%@              %oSSALC *iTemp;\n",
  "%@              if (%oSSALCfreeList==%oLambdaSSAL)\n",
  "#if defined (__cplusplus)\n",
  "                 try {\n",
  "%@                    iTemp = new %oSSALC;\n",
  "                 } catch (std::bad_alloc) {\n",
  "%@                    %ofatal (\"error in Ox Stack Solved List space "
      "allocation\\n\");\n",
  "                 }\n",
  "#else /* !defined (__cplusplus) */\n",
  "%@                 {iTemp = (%oSSALC *) malloc ((size_t) sizeof "
      "(%oSSALC));\n",
  "%@                  if (iTemp==%oLambdaSSAL)\n",
  "%@                     %ofatal (\"malloc error in Ox Stack Solved List "
      "space allocation\\n\");\n",
  "                 }\n",
  "#endif /* !defined (__cplusplus) */\n",
  "              else\n",
  "%@                 {iTemp = %oSSALCfreeList;\n",
  "%@                  %oSSALCfreeList = %oSSALCfreeList->next;\n",
  "                 }\n",
  "              iTemp->next = *pL;\n",
  "%@              iTemp->attrbNum = %oRCIL[nextP-1];\n",
  "              *pL = iTemp;\n",
  "             }\n",
  "          else\n",
  "%@             {if ((gnpDum = "
      "(LHSsemEl->node->cL)[%oRCIL[nextP-2]])->prodNum!=0)\n",
  "%@                 {long iTemp = %oIIEL[%oIIIEL[gnpDum->prodNum]] + "
      "%oRCIL[nextP-1];\n",
#if defined(DEBCTC)
  "                  (void) fprintf (stderr,\"%o\",\n",
  "                                         \"\\nsignalling ents of 0-init IN: "
      "\\n\"\n",
  "                                 );\n",
#endif
  "%@                  %oSignalEnts (gnpDum,\n",
  "%@                                 %oIEL[iTemp],\n",
  "%@                                 %oIEL[iTemp +1]\n",
  "                                );\n",
  "                 }\n",
  "             }\n",
  "         }\n",
  "      nextP += 3;\n",
  "     }\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "@@d@%@static void %oPrune (long prodNum, %oOxAttrbs *val_OxAttrbs)\n",
  "@@d@%@  {%oSIT *semEl = val_OxAttrbs->semEl;\n",
  "@@d@   int n = 0;\n",
  "@@d@   char *cp1, *cp2;\n",
  "@@d@%@   %oRCT *rcp2, *rcp3;\n",
  "@@d@   long cycleSum = 0;\n",
  "@@d@   long nNZrc = 0;\n",
  "@@d@\n",
  "@@d@   /* visit the refCountList of each node that is a child of the */\n",
  "@@d@   /* current top-of-stack node, and sum the non-zero refCounts  */\n",
  "@@d@   for (int i = 0; i < ((semEl->node)->cLlen); i++)\n",
  "@@d@%@     {if ((semEl->node)->cL[i]==(%oGNT *) NULL) continue;\n",
  "@@d@      rcp2 = ((semEl->node)->cL[i])->refCountList;\n",
  "@@d@      rcp3 = rcp2 + ((semEl->node)->cL[i])->refCountListLen;\n",
  "@@d@      while (rcp2 < rcp3)\n",
  "@@d@         if (n++, *rcp2++) {cycleSum += *(rcp2 -1); nNZrc++;}\n",
  "@@d@     }\n",
  "@@d@   if (nNZrc)\n",
  "@@d@     {\n",
  "@@C@      (void) fputs (\"\\n\\n\\n----------\\n\", stderr);\n",
  "@@C@      (void) fputs (\"cycle detected during pruning of a subtree\\n\", "
      "stderr);\n",
  "@@C@      (void) fprintf (stderr, \"  at whose root production %ld is "
      "applied.\\n\", prodNum);\n",
  "@@C@#if CYCLE_VERBOSE\n",
  "@@C@      (void) fprintf (stderr, \"number of non-root attribute instances "
      "in subtree == %d.\\n\",\n",
  "@@C@                              n\n",
  "@@C@                     );\n",
  "@@C@      (void) fprintf (stderr, \"number of unsolved attribute instances "
      "== %ld.\\n\",\n",
  "@@C@                              nNZrc\n",
  "@@C@                     );\n",
  "@@C@      (void) fprintf (stderr, \"total number of remaining dependencies "
      "== %ld.\\n\",\n",
  "@@C@                              cycleSum\n",
  "@@C@                     );\n",
  "@@C@      (void) fputs (\"average number of remaining dependencies\\n\", "
      "stderr);\n",
  "@@C@      (void) fprintf (stderr, \"  per unsolved instance == %f.\\n\",\n",
  "@@C@                              ((float) (cycleSum) / (float) (nNZrc))\n",
  "@@C@                     );\n",
  "@@C@#endif\n",
  "@@C@      (void) fprintf (stderr, \"prune aborted: searching subtree for "
      "%ld unsolved instances:\\n\",\n",
  "@@C@                              nNZrc\n",
  "@@C@                     );\n",
  "@@C@%@      %oUnsolvedInstSearchTrav (semEl->node);\n",
  "@@d@      return;\n",
  "@@d@     }\n",
  "@@d@   for (int i = 0; i < ((semEl->node)->cLlen); i++)\n",
  "@@d@%@     {if ((semEl->node)->cL[i]==(%oGNT *) NULL) continue;\n",
  "@@d@      /* free refCountList of children of the node on the top of the "
      "stack */\n",
  "@@d@      if (((semEl->node)->cL[i])->refCountList)\n",
  "@@d@#if defined (__cplusplus)\n",
  "@@d@         delete[] ((semEl->node)->cL[i])->refCountList;\n",
  "@@d@#else\n",
  "@@d@         free (((semEl->node)->cL[i])->refCountList);\n",
  "@@d@#endif\n",
  "@@d@      /* (just in case)...\n",
  "@@d@         free childList (cL) of children of the node on the top of the "
      "stack */\n",
  "@@d@      if (((semEl->node)->cL[i])->cL)\n",
  "@@d@#if defined (__cplusplus)\n",
  "@@d@         delete[] ((semEl->node)->cL[i])->cL;\n",
  "@@d@#else\n",
  "@@d@         free (((semEl->node)->cL[i])->cL);\n",
  "@@d@#endif\n",
  "@@d@      /* free childrenof the node on the top of the stack */\n",
  "@@d@#if defined (__cplusplus)\n",
  "@@d@      delete (semEl->node)->cL[i];\n",
  "@@d@#else\n",
  "@@d@      free ((semEl->node)->cL[i]);\n",
  "@@d@#endif\n",
  "@@d@     }\n",
  "@@d@   /* free the childList (cL) of the node on the top of the stack */\n",
  "@@d@   if ((semEl->node)->cL)\n",
  "@@d@     {\n",
  "@@d@#if defined (__cplusplus)\n",
  "@@d@      delete[] (semEl->node)->cL;\n",
  "@@d@#else\n",
  "@@d@      free ((semEl->node)->cL);\n",
  "@@d@#endif\n",
  "@@d@%@      (semEl->node)->cL = (%oGNT **) NULL;\n",
  "@@d@     }\n",
  "@@d@   semEl->node->cLlen = 0;\n",
  "@@d@  }\n",
  "@@d@\n",
  "@@d@\n",
  "@@d@\n",
  "%@void %oGenLeaf (int nAttrbs,\n",
  "#if defined (S_SPLINT_S)\n",
  "                /*@unused@*/\n",
  "#endif\n",
  "                __attribute__((unused))\n",
  "%@                 int typeNum, long startP, long stopP, %YSTYPE *lval)\n",
  "  {\n",
  "#if defined (lint)\n",
  "   NOTE(ARGUNUSED(typeNum))\n",
  "#endif\n",
  "%@   %oSIT **LHSsemEl = &((%oOxAttrbs *)lval)->semEl;\n",
  "#if defined (__cplusplus)\n",
  "   try {\n",
  "%@      (*LHSsemEl) = new %oSIT;\n",
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in Ox/Yacc semantic stack space "
      "allocation\\n\");\n",
  "   }\n",
  "   try {\n",
}; // skeleton9b[]

static const char* const skeleton9c[] = {
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in Ox Semantic Node space allocation\\n\");\n",
  "   }\n",
  "#else /* !defined (__cplusplus) */\n",
  "%@   (*LHSsemEl) = (%oSIT *) malloc ((size_t) sizeof (%oSIT));\n",
  "%@   if ((*LHSsemEl)==(%oSIT *) NULL)\n",
  "%@      %ofatal (\"malloc error in Ox/Yacc semantic stack space "
      "allocation\\n\");\n",
  "%@   (*LHSsemEl)->node = (%oGNT *) malloc ((size_t) sizeof (%oGNT));\n",
  "%@   if ((*LHSsemEl)->node==(%oGNT *) NULL)\n",
  "%@      %ofatal (\"malloc error in Ox Semantic Node space "
      "allocation\\n\");\n",
  "#endif /* !defined (__cplusplus) */\n",
  "%@   (*LHSsemEl)->solvedSAlist = %oLambdaSSAL;\n",
  "   (*LHSsemEl)->node->parent.stackref = (*LHSsemEl);\n",
  "   (*LHSsemEl)->node->parentIsStack = true;\n",
  "   (*LHSsemEl)->node->cLlen = 0;\n",
  "%@   (*LHSsemEl)->node->cL = (%oGNT **) NULL;\n",
  "   (*LHSsemEl)->node->refCountListLen = nAttrbs;\n",
  "   if (nAttrbs>0)\n",
  "%@     {%oRCT *rcPdum;\n",
  "#if defined (__cplusplus)\n",
  "      try {\n",
  "%@         rcPdum = (*LHSsemEl)->node->refCountList = new %oRCT "
      "[nAttrbs];\n",
  "      } catch (std::bad_alloc) {\n",
  "%@         %ofatal (\"error in Ox Reference Count List space "
      "allocation\\n\");\n",
  "      }\n",
  "#else /* !defined (__cplusplus) */\n",
  "      rcPdum = (*LHSsemEl)->node->refCountList =\n",
  "%@               (%oRCT *) calloc ((size_t) nAttrbs, (size_t) sizeof "
      "(%oRCT));\n",
  "%@      if (rcPdum==(%oRCT *) NULL)\n",
  "%@         %ofatal (\"calloc error in Ox Reference Count List space "
      "allocation\\n\");\n",
  "#endif /* !defined (__cplusplus) */\n",
  "%@      while (startP < stopP) rcPdum[%oLRCIL[startP++]] = 0;\n",
  "     }\n",
  "   else\n",
  "%@      (*LHSsemEl)->node->refCountList = (%oRCT *) NULL;\n",
  "   (*LHSsemEl)->node->prodNum = 0;\n",
  "   (*LHSsemEl)->node->whichSym = 0;\n",
#if defined(DEBCTC)
  "   (void) fprintf (stderr, \"\\ngenerating leaf node: %lx  type number "
      "%d\\n\",\n",
  "                           (unsigned long) ((*LHSsemEl)->node),\n",
  "                           typeNum\n",
  "                  );\n",
#endif
  "@@u@%@   %oLeafInstanceCount += nAttrbs; %oLeafNodeCount++;\n",
  "@@u@%@   if (!nAttrbs) %oAttrblessLeafNodeCount++;\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@void %oabort (void) {%oYok = false;}\n",
  "\n",
  "\n",
  "\n",
}; // skeleton9c[]

void
codeGen::skeletonText9 (void)
{
  putStringArray (skeleton9a, sizeof (skeleton9a));


  if (ctrl.targetParserLanguage == languageCXX)
    put ("      (*LHSsemEl)->node = new " + yoxPrefix ("GNT") +
         " (typeNum);\n");
  else
    {
      // following has preceding "#if defined (__cplusplus)" guard in skeleton
      put ("      (*LHSsemEl)->node = new " + yoxPrefix ("GNT") + ";\n");
    }

  putStringArray (skeleton9b, sizeof (skeleton9b));


  if (ctrl.targetParserLanguage == languageCXX)
    put ("      (*LHSsemEl)->node = new " + yoxPrefix ("GNT") +
         " (typeNum);\n");
  else
    {
      // following has preceding "#if defined (__cplusplus)" guard in skeleton
      put ("      (*LHSsemEl)->node = new " + yoxPrefix ("GNT") + ";\n");
    }

  putStringArray (skeleton9c, sizeof (skeleton9c));
}

// *********************************************************************

static const char* const skeleton10[] = {
  "\n",
  "\n",
  "\n",
  "%@#define %oSizeofProd(num) (%oProdsInd[(num) +1] - %oProdsInd[(num)])\n",
  "\n",
  "%@#define %oGSoccurStr(prodNum, symPos) \\\n",
  "%@   (%oStringTab[%oProds[%oProdsInd[prodNum] + (symPos)][0]])\n",
  "\n",
  "%@#define %oSizeofSort(num) (%oSortsInd[(num) +1] - %oSortsInd[(num)])\n",
  "\n",
  "%@#define %oSortOf(prodNum, symPos) \\\n",
  "%@  (%oProds[%oProdsInd[(prodNum)] + (symPos)][1])\n",
  "\n",
  "%@#define %oAttrbStr(prodNum, symPos, attrbNum)                     \\\n",
  "%@  (%oStringTab[%oSorts[%oSortsInd[%oSortOf(prodNum, symPos)] + \\\n",
  "                         (attrbNum)                                \\\n",
  "                        ]                                          \\\n",
  "               ]                                                   \\\n",
  "  )\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oShowProd (int i)\n",
  "  {int nSyms;\n",
  "\n",
  "%@   nSyms = %oSizeofProd(i);\n",
  "   for (int j = 0; j < nSyms; j++)\n",
  "     {\n",
  "%@      (void) fprintf (stderr, \"%%s\", %oGSoccurStr(i, j));\n",
  "      if (j==0) (void) fputs (\" : \", stderr); else (void) putc (\' \', "
      "stderr);\n",
  "     }\n",
  "   (void) fputs (\";\\n\", stderr);\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "@@c@%@static void %oCheckNodeInstancesSolved (%oGNT *np)\n",
  "@@c@  {int mysort, sortSize, prodNum, symPos;\n",
  "@@c@   int nUnsolvedInsts = 0;\n",
  "@@c@   bool inTerminalNode;\n",
  "@@c@\n",
  "@@c@   if (np->prodNum!=0)\n",
  "@@c@     {inTerminalNode = false;\n",
  "@@c@      prodNum = np->prodNum;\n",
  "@@c@      symPos = 0;\n",
  "@@c@     }\n",
  "@@c@   else\n",
  "@@c@     {inTerminalNode = true;\n",
  "@@c@      prodNum = np->parent.noderef->prodNum;\n",
  "@@c@      symPos = np->whichSym;\n",
  "@@c@     }\n",
  "@@c@%@   mysort = %oSortOf(prodNum, symPos);\n",
  "@@c@%@   sortSize = %oSizeofSort(mysort);\n",
  "@@c@   for (int i = 0; i < sortSize; i++)\n",
  "@@c@     if ((np->refCountList)[i]>0) nUnsolvedInsts += 1;\n",
  "@@c@   if (nUnsolvedInsts>0)\n",
  "@@c@     {(void) fprintf (stderr, \"\\nFound node that has %d unsolved "
      "attribute instance(s).\\n\",\n",
  "@@c@                              nUnsolvedInsts\n",
  "@@c@                     );\n",
  "@@c@      (void) fprintf (stderr, \"Node is labeled \\\"%s\\\".\\n\",\n",
  "@@c@%@                              %oGSoccurStr(prodNum, symPos));\n",
  "@@c@      if (inTerminalNode)\n",
  "@@c@        {(void) fputs (\"Node is terminal.  Its parent production "
      "is:\\n  \", stderr);\n",
  "@@c@%@         %oShowProd (prodNum);\n",
  "@@c@        }\n",
  "@@c@      else\n",
  "@@c@        {(void) fputs (\"Node is nonterminal.  \", stderr);\n",
  "@@c@         if (!(np->parentIsStack))\n",
  "@@c@           {(void) fprintf (stderr, \"Node is %huth child in its parent "
      "production:\\n  \",\n",
  "@@c@                                    np->whichSym\n",
  "@@c@                           );\n",
  "@@c@%@            %oShowProd (np->parent.noderef->prodNum);\n",
  "@@c@           }\n",
  "@@c@         (void) fputs (\"Node is on left hand side of this "
      "production:\\n  \", stderr);\n",
  "@@c@%@         %oShowProd (np->prodNum);\n",
  "@@c@        }\n",
  "@@c@      (void) fputs (\"The following instances are unsolved:\\n\", "
      "stderr);\n",
  "@@c@      for (int i = 0; i < sortSize; i++)\n",
  "@@c@        if ((np->refCountList)[i]>0)\n",
  "@@c@          (void) fprintf (stderr, \"     %-16s still has %1hu "
      "dependencies.\\n\",\n",
  "@@c@%@                                  %oAttrbStr(prodNum, symPos, i),\n",
  "@@c@                                  (np->refCountList)[i]\n",
  "@@c@                         );\n",
  "@@c@     }\n",
  "@@c@  }\n",
  "@@c@\n",
  "@@c@\n",
  "@@c@\n",
  "@@c@%@static void %oCheckUnsolvedInstTrav (%oGNT *pNode, long *nNZrc, long "
      "*cycleSum)\n",
  "@@c@%@  {%oGNT **cLpdum;\n",
  "@@c@%@   %oRCT *rcp;\n",
  "@@c@   int i;\n",
  "@@c@\n",
  "@@c@   /* visit the refCountList of each node in the tree, and sum the "
      "non-zero refCounts */\n",
  "@@c@   rcp = pNode->refCountList;\n",
  "@@c@   i = pNode->refCountListLen;\n",
  "@@c@   while ((i--)>0)\n",
  "@@c@      if ((*rcp++)>0) {*cycleSum += *(rcp -1); (*nNZrc)++;}\n",
  "@@c@   cLpdum = pNode->cL;\n",
  "@@c@   i = pNode->cLlen;\n",
  "@@c@   while ((i--)>0)\n",
  "@@c@     {\n",
  "@@c@%@      %oCheckUnsolvedInstTrav (*cLpdum, nNZrc, cycleSum);\n",
  "@@c@      cLpdum++;\n",
  "@@c@     }\n",
  "@@c@  }\n",
  "@@c@\n",
  "@@c@\n",
  "@@c@\n",
  "@@c@%@static void %oUnsolvedInstSearchTravAux (%oGNT *pNode)\n",
  "@@c@%@  {%oGNT **cLpdum;\n",
  "@@c@   int i;\n",
  "@@c@\n",
  "@@c@%@   %oCheckNodeInstancesSolved (pNode);\n",
  "@@c@   cLpdum = pNode->cL;\n",
  "@@c@   i = pNode->cLlen;\n",
  "@@c@   while ((i--)>0)\n",
  "@@c@     {\n",
  "@@c@%@      %oUnsolvedInstSearchTravAux (*cLpdum);\n",
  "@@c@      cLpdum++;\n",
  "@@c@     }\n",
  "@@c@  }\n",
  "@@c@\n",
  "@@c@\n",
  "@@c@\n",
  "@@C@%@static void %oUnsolvedInstSearchTrav (%oGNT *pNode)\n",
  "@@C@%@  {%oGNT **cLpdum;\n",
  "@@C@   int i;\n",
  "@@C@\n",
  "@@C@   cLpdum = pNode->cL;\n",
  "@@C@   i = pNode->cLlen;\n",
  "@@C@   while ((i--)>0)\n",
  "@@C@     {\n",
  "@@C@%@      %oUnsolvedInstSearchTravAux (*cLpdum);\n",
  "@@C@      cLpdum++;\n",
  "@@C@     }\n",
  "@@C@  }\n",
  "@@C@\n",
  "@@C@\n",
  "@@C@\n",
}; // skeleton10[]

void
codeGen::skeletonText10 (void)
{
  putStringArray (skeleton10, sizeof (skeleton10));
}

// *********************************************************************

static const char* const skeletonRRsection[] = {
  "%@static void %oExecuteRRsection (%oGNT *rootNode)\n",
  "  {\n",
  "@@c@   long cycleSum = 0;\n",
  "@@c@   long nNZrc = 0;\n",
  "\n",
#if defined(DEBCTC)
  "%@   (void) fputs (\"\\n\\n\\nin %oExecuteRRsection \\n\", stderr);\n",
  "   (void) fputs (\"\\n/******************************************/ \\n\", "
      "stderr);\n",
#endif
  "%@   if (!%oYok) return;\n",
  "@@u@(void) fprintf (stderr, \"Size of input: %8ld tokens\\n\\n\",\n",
  "@@u@%@                        %oLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) putc (\'\\n\', stderr);\n",
  "@@u@(void) fprintf (stderr, \"%8ld attribute instances in %8ld leaf nodes "
      "\\n\",\n",
  "@@u@%@                        %oLeafInstanceCount, %oLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fprintf (stderr, \"%8ld attribute instances in %8ld interior "
      "nodes \\n\",\n",
  "@@u@%@                        %oIntInstanceCount, %oIntNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fprintf (stderr, \"%8ld attribute instances in %8ld parse-tree "
      "nodes\\n\",\n",
  "@@u@%@                        %oIntInstanceCount + %oLeafInstanceCount,\n",
  "@@u@%@                        %oIntNodeCount + %oLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fprintf (stderr, \"%8ld attribute instances solved during "
      "decorations\\n\",\n",
  "@@u@%@                        %oDecSolvedInstCount\n",
  "@@u@               );\n",
  "@@u@%@if (%oIntInstanceCount + %oLeafInstanceCount)\n",
  "@@u@   (void) fprintf (stderr, \"        %8ld/%ld = %.3f \\n\",\n",
  "@@u@%@                           %oDecSolvedInstCount,\n",
  "@@u@%@                           %oIntInstanceCount + "
      "%oLeafInstanceCount,\n",
  "@@u@%@                           ((float) %oDecSolvedInstCount) /\n",
  "@@u@%@                           ((float) (%oIntInstanceCount + "
      "%oLeafInstanceCount))\n",
  "@@u@                  );\n",
  "@@u@(void) fprintf (stderr, \"%8ld attributeless leaf nodes \\n\",\n",
  "@@u@%@                        %oAttrblessLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) putc (\'\\n\', stderr);\n",
  "@@u@(void) fprintf (stderr, \"# parse-tree nodes / # leaf nodes = %5.2f "
      "\\n\",\n",
  "@@u@%@                        (%oIntNodeCount + %oLeafNodeCount) /\n",
  "@@u@%@                        (float) %oLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fputs (\"# leaf attribute instances / \", stderr);\n",
  "@@u@(void) fprintf (stderr, \"# leaf nodes = %5.2f\\n\",\n",
  "@@u@%@                        %oLeafInstanceCount /\n",
  "@@u@%@                        (float) %oLeafNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fputs (\"# interior attribute instances / \", stderr);\n",
  "@@u@(void) fprintf (stderr, \"# interior nodes = %5.2f\\n\",\n",
  "@@u@%@                        %oIntInstanceCount /\n",
  "@@u@%@                        (float) %dIntNodeCount\n",
  "@@u@               );\n",
  "@@u@(void) fputs (\"total # attribute instances / \", stderr);\n",
  "@@u@(void) fprintf (stderr, \"# parse-tree nodes = %5.2f\\n\",\n",
  "@@u@%@                        (%oIntInstanceCount + %oLeafInstanceCount) "
      "/\n",
  "@@u@%@                        (float) (%oIntNodeCount + %oLeafNodeCount)\n",
  "@@u@               );\n",
  "@@c@%@   %oCheckUnsolvedInstTrav (rootNode, &nNZrc, &cycleSum);\n",
  "@@c@   if (nNZrc>0)\n",
  "@@c@      {\n",
  "@@c@       (void) fputs (\"\\n\\n\\n**********\\n\", stderr);\n",
  "@@c@       (void) fputs (\"cycle detected in completed parse tree\", "
      "stderr);\n",
  "@@c@       (void) fputs (\" after decoration.\\n\", stderr);\n",
  "@@c@#if CYCLE_VERBOSE\n",
  "@@c@       (void) fprintf (stderr, \"number of unsolved attribute instances "
      "== %ld.\\n\",\n",
  "@@c@                               nNZrc\n",
  "@@c@                      );\n",
  "@@c@       (void) fprintf (stderr, \"total number of remaining dependencies "
      "== %ld.\\n\",\n",
  "@@c@                               cycleSum\n",
  "@@c@                      );\n",
  "@@c@       (void) fputs (\"average number of remaining dependencies\\n\", "
      "stderr);\n",
  "@@c@       (void) fprintf (stderr, \"  per unsolved instance == %f.\\n\",\n",
  "@@c@                               ((float) (cycleSum) / (float) (nNZrc))\n",
  "@@c@                      );\n",
  "@@c@#endif\n",
  "@@c@       (void) fprintf (stderr, \"searching parse tree for %ld unsolved "
      "instances:\\n\",\n",
  "@@c@                               nNZrc\n",
  "@@c@                      );\n",
  "@@c@%@       %oUnsolvedInstSearchTravAux (rootNode);\n",
  "@@c@      }\n",
  "%@  } /* %oExecuteRRsection */ \n",
}; // skeletonRRsection[]

void
codeGen::putRootAttrbRefSection ()
{
  putStringArray (skeletonRRsection, sizeof (skeletonRRsection));
}

// *********************************************************************

static const char* const skeletonSASHeaderA[] = {
  "@@u@\n",
  "@@u@\n",
  "@@u@\n",
  "@@u@%@static long %oIntInstanceCount = 0; static long %oIntNodeCount = 0;\n",
  "@@u@%@static long %oLeafInstanceCount = 0; static long %oLeafNodeCount = ",
                                                                     "@@u@0;\n",
  "@@u@%@static long %oAttrblessLeafNodeCount = 0;\n",
  "@@u@%@static long %oDecSolvedInstCount = 0;\n",
  "\n",
  "\n",
  "\n",
}; // skeletonSASHeaderA[]

static const char* const skeletonSASHeaderB[] = {
  "  {\n",
  "\n",
  "\n",
  "\n",
  "%@#define %oDECORfREQ 50\n",
  "\n",
  "\n",
  "\n",
  "%@   static unsigned short intNodeCount = %oDECORfREQ;\n",
  "   if (--intNodeCount==0)\n",
  "%@      intNodeCount = %oDECORfREQ;\n",
  "   else\n",
  "      if (!pruning) return;\n",
#if defined(DEBCTC)
  "   (void) fputs (\"\\n\\n\\ndecorating\\n\", stderr);\n",
  "   (void) fputs (\"/******************************************/\\n\", "
      "stderr);\n",
#endif
  "%@   while (%oRSTop >= %oRSBottom)\n",
  "     {\n",
  "%@%oWST whichSym; %oWAT whichAttrb;\n",
  "%@%oGNT *RSTopNode, *RSNode;\n",
  "%@%oParent RSTopParent;\n",
  "\n",
  "\n",
  "%@RSTopParent = (RSTopNode = %oRS[%oRSTop].node)->parent;\n",
  "%@RSNode = ((whichSym = (%oRS[%oRSTop].whichSym))!=0) ? RSTopParent.noderef",
                                                              " : RSTopNode;\n",
#if defined(DEBCTC)
  "\n",
  "%@(void) fputs (\"\\nentering %oSolveAndSignal switch: \\n\", stderr);\n",
  "\n",
#endif
  "%@whichAttrb = %oRS[%oRSTop].whichAttrb;\n",
  "%@%oRSTop--;\n",
  "switch (RSNode->prodNum) {\n",
}; // skeletonSASHeaderB[]

void
codeGen::putSolveAndSignalHeader (void)
{
  putStringArray (skeletonSASHeaderA, sizeof (skeletonSASHeaderA));

  putDecoratePrototype (); put ("\n");

  putStringArray (skeletonSASHeaderB, sizeof (skeletonSASHeaderB));
}

// *********************************************************************

static const char* const skeletonSASTrailer[] = {
  "} /* switch */\n",
  "\n",
  "if (whichSym!=0) /* the just-solved instance was inherited. */\n",
  "   {if (RSTopNode->prodNum!=0)\n",
  "%@       {long iDum = %oIIEL[%oIIIEL[RSTopNode->prodNum]] + whichAttrb;\n",
#if defined(DEBCTC)
  "        (void) fprintf (stderr, \"%o\",\n",
  "%@                                \"\\nsignalling ents of attrb in "
      "%oSolveAndSignal trailer: \\n\"\n",
  "                       );\n",
#endif
  "%@        %oSignalEnts (RSTopNode, %oIEL[iDum],\n",
  "%@                                  %oIEL[iDum +1]\n",
  "                      );\n",
  "       }\n",
  "   }\n",
  "else     /* the just-solved instance was synthesized. */\n",
  "   {if (!(RSTopNode->parentIsStack)) /* node has a parent. */\n",
  "%@       {long iDum = %oIIEL[%oIIIEL[RSTopParent.noderef->prodNum] +\n",
  "                          RSTopNode->whichSym\n",
  "                         ] +\n",
  "                  whichAttrb;\n",
#if defined(DEBCTC)
  "        (void) fprintf (stderr, \"%o\",\n",
  "%@                                \"\\nsignalling ents of attrb in "
      "%oSolveAndSignal trailer: \\n\"\n",
  "                       );\n",
#endif
  "%@        %oSignalEnts (RSTopParent.noderef,\n",
  "%@                       %oIEL[iDum],\n",
  "%@                       %oIEL[iDum +1]\n",
  "                      );\n",
  "       }\n",
  "    else   /* node is still on the stack--it has no parent yet. */\n",
  "%@       {%oSSALC **pL = &(RSTopParent.stackref->solvedSAlist);\n",
  "%@        %oSSALC *iDum;\n",
  "%@        if (%oSSALCfreeList==%oLambdaSSAL)\n",
  "#if defined (__cplusplus)\n",
  "            try {\n",
  "%@               iDum = new %oSSALC;\n",
  "            } catch (std::bad_alloc) {\n",
  "%@               %ofatal (\"error in Ox Stack Solved List space "
      "allocation\\n\");\n",
  "            }\n",
  "#else\n",
  "%@           {iDum = (%oSSALC *) malloc ((size_t) sizeof (%oSSALC));\n",
  "%@            if (iDum==%oLambdaSSAL)\n",
  "%@               %ofatal (\"malloc error in Ox Stack Solved List space "
      "allocation\\n\");\n",
  "           }\n",
  "#endif\n",
  "        else\n",
  "%@           {iDum = %oSSALCfreeList;\n",
  "%@            %oSSALCfreeList = %oSSALCfreeList->next;\n",
  "           }\n",
  "        iDum->next = *pL;\n",
  "%@        iDum->attrbNum = (%oWAT) whichAttrb;\n",
  "        *pL = iDum;\n",
  "       }\n",
  "   }\n",
  "\n",
  "@@u@      %oDecSolvedInstCount++;\n",
  "     }\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
}; // skeletonSASTrailer[]

void
codeGen::putSolveAndSignalTrailer ()
{
  if (genRSNodeProdNumSwitchHd)
    putStringArray (skeletonSASTrailer, sizeof (skeletonSASTrailer));
  else
    {
      putStringArray (skeletonSASHeaderA, sizeof (skeletonSASHeaderA));

      putDecoratePrototype (); put ("\n"
                                    "  { /* no-op */ }"
                                    "\n\n\n\n");
    }
}

// protected:
// private:
// ================================ end codeGen ================================
