/* $Id: mac.cc 2839 2022-12-13 00:53:33Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <sstream>
#include <stack>
#include <unordered_map>

#include "location.hh"

#include "mac.hh"
#include "messages.hh"
#include "outMgr.hh"
#include "scan.hh"



int
macroDef::macroFormalLookup (const std::string& cp) const
{
  const int numFormals = (int)formals.size ();

  for (int i = 0; i < numFormals; i++)
    if (formals[i] == cp)
      return (i);

  return -1;
}



bool
macroDef::installMacroFormal (const std::string& cp)
{
  if (macroFormalLookup (cp) == -1)
    {
      // new formal parameter: add it to the formal parameter list

      try
        {
          formals.push_back (cp);
        }
      catch (...)
        {
          // expansion of the macro definition formal parameter list failed
          yox_alloc_fail (__FILE__, __LINE__);
        }

      return true;
    }
  else
    {
      // duplicated formal parameter

      return false;
    }
}



typedef std::unordered_map<std::string, macroDef> macroTableType;

static macroTableType macroTable; // initially empty



void
macroDef::close (void)
{
  flush (); // write any currently buffered text to the internal memory file

  body = getObj ().str (); // the internal memory file is the body

  try
    {
      // not a problem if inserting a duplicate: already issued an error message
      (void)macroTable.insert (macroTableType::value_type (name, *this));
    }
  catch (...)
    {
      // insert on the macro table failed
      yox_alloc_fail (__FILE__, __LINE__);
    }
}



macroDef macro;



struct actual
{
  std::string arg;
  AGC::position pos;

  actual (std::string a, AGC::position& p)
      : arg (a),
        pos (p)
  {}
};

typedef std::vector<struct actual> actualList;



class macroExpansion
{
public:

  macroExpansion (const std::string& n, const AGC::position from)
      : name (n),
        where_invoked (from),
        inputFromBody (false),
        inputFromActual (false)
  {}

  const std::string name;
  const AGC::position where_invoked;
  bool inputFromBody;
  AGC::location substLoc;
  bool inputFromActual;
  actualList actuals;
};



// nested macro expansion stack
static std::stack<macroExpansion> macroStack;



const macroDef&
getCurMacro (void)
{
  assert (expandingMacro ());
  return macroTable[macroStack.top ().name]; // entry guaranteed to be present
}



bool
macroLookup (const std::string& cp)
{
  try
    {
      (void)macroTable.at (cp);
      return true;
    }
  catch (std::out_of_range const&) // not present
    {
      return false;
    }
}



void
macroArg::terminate (void)
{
  flush (); // write any currently buffered text to the internal memory file

  // the internal memory file holds the argument
  macroStack.top ().actuals.push_back (actual (getObj ().str (), origin));

  clear (); // clear the parameter
}



void
expandMacro (const std::string& name, const AGC::position where)
{
  macroStack.push (macroExpansion (name, where));
}



void
setupMacroExpansion (AGC::location const& loc,
                     std::string& body,
                     AGC::position& origin)
{
  macroDef& mdDum = macroTable[macroStack.top ().name]; // the entry exists

  const int numFormals = (int)mdDum.formals.size ();

  const int numActuals = (int)macroStack.top ().actuals.size ();

  if (numActuals > numFormals)
    {
      // not currently fatal - should it be?

      yox_error (loc, "too many arguments in macro invocation.\n");
    }


  if (numActuals < numFormals)
    yox_fatal (loc, "too few arguments in macro invocation.\n");

  macroStack.top ().inputFromBody = true;

  body = mdDum.body;
  origin = mdDum.getOrigin ();
}



bool
expandingMacro (void)
{
  if (macroStack.empty ())
    return false;

  return macroStack.top ().inputFromBody;
}



void
macroDone (void)
{
  if (macroStack.empty ())
    return;

  macroStack.top ().inputFromBody = false;
  macroStack.pop ();
}



// get the actual parameter corresponding to formal parameter 'argIndex'
const std::string&
substFormal (int argIndex)
{
  return macroStack.top ().actuals[argIndex].arg;
}



void
expandFormalsInAttrOccr (std::string& attrb_occur)
{
  bool replace = false;

  const macroDef& theMacro (getCurMacro ());

  std::string gsymName, occurrence, attrName;

  getAttrRefComponents (attrb_occur, gsymName, occurrence, attrName);

  int argIndex;
  std::string substValue;

  argIndex = theMacro.macroFormalLookup (trim (gsymName));
  if (argIndex != -1)
    {
      replace = true;
      substValue.append (substFormal (argIndex));
    }
  else
    substValue.append (gsymName);

  substValue.append (".");

  if (!occurrence.empty ())
    {
      argIndex = theMacro.macroFormalLookup (trim (occurrence));
      if (argIndex != -1)
        {
          replace = true;
          substValue.append (substFormal (argIndex));
        }
      else
        substValue.append (occurrence);

      substValue.append (".");
    }
  else
    substValue.append ("0.");

  argIndex = theMacro.macroFormalLookup (trim (attrName));
  if (argIndex != -1)
    {
      replace = true;
      substValue.append (substFormal (argIndex));
    }
  else
    substValue.append (attrName);

  if (replace)
    attrb_occur = substValue;
}



void
setupFormalExpansion (const AGC::location& loc)
{
  if (macroStack.empty ())
    return;

  macroStack.top ().inputFromActual = true;
  macroStack.top ().substLoc = loc;
}

bool
replacingFormal (void)
{
  if (macroStack.empty ())
    return false;

  return macroStack.top ().inputFromActual;
}

void
macroFormalDone (void)
{
  if (macroStack.empty ())
    return;

  macroStack.top ().inputFromActual = false;
}



const std::string&
macroName (void)
{
  return macroStack.top ().name;
}

int
macroInvokedFrom (void)
{
  return macroStack.top ().where_invoked.line;
}

const AGC::location&
formalLoc (void)
{
  return macroStack.top ().substLoc;
}



bool
initMacroDef (const std::string& cp)
{
  if (macroLookup (cp)) // duplicated macro name
    {
      macro = macroDef (macroTable[cp]); // the entry exists

      macro.formals.clear (); // reset the formal parameter list
      return false;
    }
  else
    {
      macro = macroDef (cp);
      return true;
    }
}
