/* $Id: opts.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef OPTS_HH_INCLUDED
#define OPTS_HH_INCLUDED

#include <string>

#include "location.hh"



enum whichYoxLexer
{
  NO_LEXER,
  YACC_SPEC_LEXER,
  LEX_SPEC_LEXER,
  GEN_SPEC_LEXER
};


// supported lexer generators
enum lexerGenKind
{
  flexLexerGen   = 1,
  reflexLexerGen = 2,
  lexLexerGen    = 0 // default
};


// supported parser generators
enum parserGenKind
{
  bisonParserGen    = 1,
  bisonParserGenOld = 2,
  byaccParserGen    = 4,
  btyaccParserGen   = 8,
  mstaParserGen     = 16,
  yaccParserGen     = 0 // default
};


// supported languages
enum languageKind
{
  languageC       = 1,
  languageCXX     = 2,
  languageUnknown = 0 // default
};


class yoxControl
{
public:

  yoxControl ()
      : process_autoSynOrInh (false),

        targetParserGen (yaccParserGen),
        targetParserLanguage (languageC),

        targetLexerGen (lexLexerGen),
        targetLexerLanguage (languageC),

        targetLexerGenForThisLfile (lexLexerGen),
        targetLexerLanguageForThisLfile (languageC),

        allLfilesForReentrantParser (false),
        thisLfileForReentrantParser (false),

        buildHeaderFile (false),

        genTreeViz (false),

        neutralizeSemActions (false),

#if WITH_LINE_DIRECTIVES
        noLineDirectives (false), // generate `#line' directives
#else
        noLineDirectives (true), // suppress `#line' directives
#endif

        saveBadOutputs (false),

        showPTsizeStats (false),

        stripOxConstructs (false),

        suppressCycleReports (false),

        suppressPruning (false),

        removeOutput (false),

        yaccSymbolPrefix ("yy"),
        yoxSymbolPrefix ("yyy"),
        user_namespace ("yy"),
        user_parser ("parser"),
        user_token_prefix (""),
        unionTag ("")
  {}

  // Y-file lexer hack: set true in parser only while processing global
  // automatic copy rule attribute modifier declarations @autoinh & @autosyn;
  // suppresses copying of the attribute names declared to the output file
  bool process_autoSynOrInh;

  // set to a specific Yacc-compatible parser generator if specified by a
  // command line option
  enum parserGenKind targetParserGen;

  // set to `languageCXX' if the '--bison=c++', '--btyacc=c++' or '--msta=c++'
  // option is found on the command line, or if it can be inferred from options
  // or directives in the Y-file
  enum languageKind targetParserLanguage;

  // set to a specific Lex-compatible lexer generator if specified by a command
  // line option
  enum lexerGenKind targetLexerGen;

  // set to `languageCXX' if '--flex=c++' option is found on the command line
  enum languageKind targetLexerLanguage;

  // set to a specific Lex-compatible lexer generator if it can be inferred from
  // options or directives in the L-file
  enum lexerGenKind targetLexerGenForThisLfile;

  // set to `languageCXX' if '%option c++' is found in the L-file;
  enum languageKind targetLexerLanguageForThisLfile;

  // set to `true' if '--pure-parser' or '--c++=bison' option (or equivalent)
  // found on the command line, or if '%pure-parser' or '%language "C++"' (or
  // equivalent) is found in the Y-file
  bool allLfilesForReentrantParser;

  // set to `true' if '%option bison-bridge' or '%option bison-locations' found
  // in the lexer definition file
  bool thisLfileForReentrantParser;

  bool buildHeaderFile;

  bool genTreeViz;

  bool neutralizeSemActions;

  bool noLineDirectives;

  bool saveBadOutputs;

  // set to 'true' to generate an evaluator that reports maximum sizes of some
  // parse-tree-related data structures upon termination.
  bool showPTsizeStats;

  bool stripOxConstructs;

  bool suppressCycleReports;

  bool suppressPruning;

  bool removeOutput;

  std::string yaccSymbolPrefix;
  std::string yoxSymbolPrefix;

  // 'bison' extensions
  std::string user_namespace;    // user-specified
  std::string user_parser;       // user-specified class name
  std::string user_token_prefix; // user-specified
  std::string unionTag;          // optional tag following `%union'
};



extern const std::string yaccPrefix (const std::string, bool uc = false);
extern const std::string yoxPrefix (const std::string);
extern const std::string addTokenPrefix (const std::string);
extern const std::string stripTokenPrefix (const std::string);

// check the Flex & RE/flex '%option' directive for those of interest to Ox
extern void checkLexGenOptionDirective (const AGC::location&, std::string);
// set '*ForThisLexer' inferences for RE/flex unique directives
extern void reflexDirective (void);

extern void setOptions (int, char* const[]);

extern void openIOfiles (void);

extern void closeIOfiles (void);
extern void eraseIOfiles (void);

#endif // !OPTS_HH_INCLUDED
