/* $Id: outMgr.cc 2837 2022-12-11 19:12:25Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <string>

#include "outMgr.hh"
#include "messages.hh"
#include "scan.hh"



inline static void
putCharInBuffer (char c, std::string& buffer)
{
  try
    {
      buffer.push_back (c);
    }
  catch (...)
    {
      // expansion of the output buffer failed
      yox_alloc_fail (__FILE__, __LINE__);
    }
}

inline static void
putStringInBuffer (const std::string& s, std::string& buffer)
{
  try
    {
      buffer.append (s);
    }
  catch (...)
    {
      // expansion of the output buffer failed
      yox_alloc_fail (__FILE__, __LINE__);
    }
}


// =============================== class outMgr ================================

// public:

// clear the output object
void
outMgr::clear (void)
{
  name.clear ();
  outLineno = 1;
  atBOL = true;
  lastInlineno = -1;
  whitespace.clear ();
  buffer.clear ();
  blankLines = 0;
}



// is the output object empty?
bool
outMgr::empty (void)
{
  flush (); // write any the held & buffered text to the output object

  return (tellp () == 0l);
}



outMgr&
outMgr::put (const std::string& str)
{
  for (const char& c : str)
    put (c);
  return *this;
}



outMgr&
outMgr::put (int i, int width)
{
  std::string repr (std::to_string (i));

  int padding = width - repr.size ();
  if (padding > 0)
    put (std::string (padding, ' '));

  put (repr);

  return *this;
}



outMgr&
outMgr::put (long l, int width)
{
  std::string repr (std::to_string (l));

  int padding = width - repr.size ();
  if (padding > 0)
    put (std::string (padding, ' '));

  put (repr);

  return *this;
}



outMgr&
outMgr::lex (void)
{
  if (!ctrl.noLineDirectives)
    putInLineDirective (curPos->line, *curPos->filename);
  return *this;
}



outMgr&
outMgr::gen (void)
{
  if (!ctrl.noLineDirectives)
    putOutLineDirective ();
  return *this;
}


// protected:

void
outMgr::putCharInWriteBuffer (char c)
{
  if (std::isspace (c))
    {
      if (c == '\n')
        {
          if (!whitespace.empty ())
            {
              // remove any whitespace immediately preceding the EOL
              whitespace.clear ();
            }

          if (buffer.empty ()) // this is a blank line
            {
              // don't automatically write blank lines, just count them
              ++blankLines;
            }
          else
            putCharInBuffer (c, buffer);

          if (!buffer.empty ())
            {
              // the held text is a non-blank text line (ends with '\n')

              if (flushBlankLines ())
                {
                  // only need a #line directive if we skipped some input lines

                  if (!ctrl.noLineDirectives && !wasOutName)
                    {
                      // put out an input file #line directive before
                      // flushing the output line
                      std::string lineDir = lineDirective (lastInlineno,
                                                           *curPos->filename,
                                                           lexerOutput);
                      lineDir.push_back ('\n');
                      // write the #line directive out immediately
                      getObj ().write (lineDir.c_str (), lineDir.size ());
                      ++outLineno;
                    }
                }

              flush (); // write the held & buffered text to the output object
            }

          if (lastInlineno > 0)
            ++lastInlineno;

          atBOL = true;
        }
      else
        putCharInBuffer (c, whitespace);
    }
  else
    {
      if (!whitespace.empty ())
        {
          // whitespace precedes this non-whitespace character:
          // send it to the output object
          putStringInBuffer (whitespace, buffer);
          whitespace.clear (); // clear the whitespace just output
        }

      putCharInBuffer (c, buffer);

      atBOL = false;
    }
}



#define outBufSz 4 // max number of consecutive blank lines



// flush any pending blank lines; return true if some were skipped
bool
outMgr::flushBlankLines (void)
{
  if (blankLines > 0)
    {
      bool skipBlankLines = blankLines > outBufSz;
      int numBlankLines = skipBlankLines ? outBufSz : blankLines;
      blankLines = 0;

      // write out the queued blank lines immediately
      for (int i = 0; i < numBlankLines; i++)
        {
          getObj ().write ("\n", 1);
          ++outLineno;
        }

      return skipBlankLines;
    }

  return false;
}



// write any held & buffered text to the output object
void
outMgr::flush (void)
{
  // flush any pending blank lines
  flushBlankLines ();

  // write the held text to the output object
  if (!buffer.empty ())
    {
      getObj ().write (buffer.c_str (), buffer.size ());
      buffer.clear (); // clear the output line
      ++outLineno;
    }

  // write any the held whitespace to the output object
  if (!whitespace.empty ())
    {
      getObj ().write (whitespace.c_str (), whitespace.size ());
      whitespace.clear (); // clear the whitespace just output
    }

  getObj ().flush (); // write any buffered text to the output object
}



std::string
outMgr::lineDirective (int lineno, const std::string& fileName, bool forLexer)
{
  std::string line_directive;

  if (forLexer)
    line_directive.push_back (' ');

  line_directive.append ("#line ")
      .append (std::to_string (lineno))
      .append (" \"").append (fileName).append ("\"");

  return line_directive;
}



void
outMgr::putLineDirective (const int lineNo,
                          const std::string& fileName,
                          bool forLexer)
{
  put (lineDirective (lineNo, fileName, forLexer));
}



void
outMgr::putOutLineDirective (void)
{
  assert (!ctrl.noLineDirectives);

  if (!wasOutName)
    {
      if (!atBOL)
        put ('\n');

      whitespace.clear (); // remove any saved whitespace

      flushBlankLines (); // so that outLineno is correct

      putLineDirective (outLineno + 1, name, lexerOutput);
      put ('\n');

      wasOutName = true;
      lastInlineno = -1;
    }
}



void
outMgr::putInLineDirective (int inlineno, const std::string& inName)
{
  assert (!ctrl.noLineDirectives);

  if (inlineno == -1)
    return;

  if (wasOutName || (lastInlineno > 0 && lastInlineno != inlineno))
    {
      if (!atBOL)
        put ('\n');

      whitespace.clear (); // remove any saved whitespace

      flushBlankLines (); // so that outLineno is correct

      putLineDirective (inlineno, inName, lexerOutput);
      put ('\n');

      wasOutName = false;
      lastInlineno = inlineno;
    }
}



void
outMgr::append (std::iostream& from, std::string& fromName)
{
  for (std::string xferBuf; std::getline (from, xferBuf);)
    {
      long end_lndir;
      std::string fileName;

      if (!xferBuf.empty () && xferBuf[0] == '#' &&
          ((end_lndir = getLineDir (xferBuf.c_str (), fileName)) > 0))
        {
          // this is a #line directive

          if (fileName == fromName)
            {
              // change the #line directive to reference the target filename

              flushBlankLines (); // so that outLineno is correct
              putLineDirective (outLineno + 1, name, false);

              wasOutName = true;
              if (++end_lndir < (long)xferBuf.length ())
                {
                  // chars follow the matched #line directive
                  put (xferBuf.substr (end_lndir));
                }
            }
          else
            {
              // passthru the #line directive

              put (xferBuf);
              wasOutName = false;
            }
        }
      else
        {
          // this is a normal line of text

          put (xferBuf);
        }

      if (!from.eof ())
        put ('\n'); // std::getline () drops the '\n' delimiter
    }
}


// private:

// ================================ end outMgr =================================
