%state LEX_DEFS RULES_PROLOG
%state LEX_REGEXP LEX_ACT
%state LEX_CCODE CCODE_BLOCK
%state LEX_USER LEX_TRAILER
%state GEN_INITIAL GEN_ACT
%state COMMENT
%state STRING
%state CPP_DIRECTIVE

%state EXPAND_MACRO MACRO_ARG
%state LEX_RTN_OBJS GEN_RTN_OBJS
%state LEX_IS_ATTR_REF GEN_IS_ATTR_REF
%state LEX_ATTR_REF GEN_ATTR_REF

%option c++
%option namespace=AGC::LFILE lexer=Lfile_Lexer lex=yylex
%option bison-cc bison-locations
%option bison-cc-namespace=AGC bison-cc-parser=lexer
%option YYSTYPE=AGC::lexer::value_type

%{
/* $Id: agc-Lfile-lexer.ll 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <string>

#include "codegen.hh"
#include "lextypes.hh"
#include "symTrans.hh"
%}

%top{
#if CAN_SET_WITH_REALLOC
#define WITH_REALLOC 0
#endif

#include <fstream>
#include <stack>

#include "agc-Lfile.hh"
#include "mac.hh"
#include "macutil.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
}

%class{
protected:

  inline void
  push_state (const position& pos, int new_state)
  {
    AbstractLexer::push_state (new_state);
    startPos.push (pos); // starting position if EOF in 'new_state'
  }

  inline void
  push_state (const location& loc, int new_state)
  {
    AbstractLexer::push_state (new_state);
    startPos.push (loc.begin); // starting position if EOF in 'new_state'
  }

  inline void
  pop_state (void)
  {
    AbstractLexer::pop_state ();
    startPos.pop ();
  }

  Lfile_specIO& theFiles;

  EOLctx EOLstate;

private:

  int nLbraces; // '{' ... '}' nesting
  int nLcurls; // "%{" ... "%}" flex action bracket extension
  int nLparens; // '(' ... ')' nesting
  int nLparensM; // macro invocation '(' ... ')' nesting

  outMgr* whereTo;
  void
  yox_push_state (int, outMgr&);

  std::stack<position> startPos;

  std::stack<EOLctx> EOLstateStk;

  bool stringHasBareNewline;

  macroArg macroARG;
}

// L-file lexer constructor specification
%option ctorarg="Lfile_specIO& fR"
%option ctorinit="theFiles (fR), EOLstate (EOLctx (position (&fR.inName, 0, 0)))"
%init{
    nLbraces = 0;
    nLcurls = 0;
    nLparens = 0;
    nLparensM = 0;

    switch (theFiles.scanner)
      {
      case LEX_SPEC_LEXER:
        start (LEX_DEFS);
        break;
      case GEN_SPEC_LEXER:
        start (GEN_INITIAL);
        break;
      default:
        break;
      }
}

%{
#define yyless matcher ().less
#define yyinput matcher ().input
%}

WS		[ \t\f\v]+
OPTWS		[ \t\f\v]*
NON_WS		[^ \t\f\v\n]

ESCSEQ		\\([^\n]|[0-7]{1,3}|x[[:xdigit:]]{1,2})
CHAR_LIT	"'"([^\\'\n]|{ESCSEQ})"'"
STRING_LIT	"\""([^\\"\n]|{ESCSEQ})*"\""

CHAR_CONST	"L"?{CHAR_LIT}
STRING_CONST	"L"?{STRING_LIT}

ID		[a-zA-Z_][a-zA-Z_0-9]*
DOT_NAME	[_\.]*[a-zA-Z][-_\.a-zA-Z0-9]*

INTEGER		[0-9]+
HEX		"0"[xX][0-9a-fA-F]+
OCTAL		"0"[0-7]+
INT_SUFFIX	([uU][lL]{0,2})|([lL]{1,2}[uU]?)
NUMBER		({INTEGER}|{HEX}|{OCTAL}){INT_SUFFIX}?

SYMBOL		{DOT_NAME}|{CHAR_LIT}
ATTRB_OCCUR	{SYMBOL}{OPTWS}"."({OPTWS}{INTEGER}(OPTWS)".")?{OPTWS}{ID}
ATTRB_REF	"@"{OPTWS}{ATTRB_OCCUR}({OPTWS}"@")?

START_CONDS	"<"{ID}(","{ID})*">"

PAT_NAME	[_a-zA-Z][-_a-zA-Z0-9]*
NAMED_PAT	"{"{PAT_NAME}"}"

DIFF_OR_UNION	"{-}"|"{+}"

CHARCLASS_PAT	"["([^\\\]\n]|{ESCSEQ})+"]"
OTHER		([^<*+?{}^$" [\]\t\f\v\n][^*+?{}^$" [\]\t\f\v\n]?)|{DIFF_OR_UNION}
REPEAT		\*|\+|\?|"{"{INTEGER}(","{INTEGER}?)?"}"
L_ANCHOR	"^"
R_ANCHOR	"$"
COMPONENT	({STRING_LIT}|{ESCSEQ}|{CHARCLASS_PAT}|{NAMED_PAT}|{OTHER}){REPEAT}?
PATTERN		{L_ANCHOR}?{COMPONENT}+{R_ANCHOR}?

CPP_DIRECTIVE	^{OPTWS}"#"{OPTWS}("define"|"if"|"ifdef"|"ifndef"|"else"|"elif"|"endif").*$

%%

<GEN_INITIAL>{

\n {
  yoxOutL.lex ().put ((char)chr ()); // avoid 'yyless (0)' if EOL
  start (GEN_ACT);
  // hook for Yacc action at start of a generic scanner file
  return (lexer::token::SOSF);
}

. {
  yyless (0);
  start (GEN_ACT);
  // hook for Yacc action at start of a generic scanner file
  return (lexer::token::SOSF);
}

}



<LEX_DEFS>{

  /* L-file Definitions Section */

^"%{".*$ {
  yoxOutL.lex ().put (text ());
  push_state (yylloc, LEX_CCODE);
}

"%bison-bridge" {
  reflexDirective (); // RE/flex unique option = Flex option
                      // "%option bison-bridge"
}

"%bison-complete" {
  reflexDirective (); // RE/flex unique option requires Bison directives
                      // "%define api.value.type variant" and
                      // "%define api.token.constructor", which
                      // are not supported
  notSupported (yylloc, "\"%bison-complete\"");
}

"%bison-locations" {
  reflexDirective (); // RE/flex unique option = Flex "%option bison-locations"
}

"%class"{OPTWS}"=" {
  // disambiguate with "%class{ ... }"
  reflexDirective (); // RE/flex unique option
}

"%class" {
  reflexDirective (); // RE/flex unique option
  yoxOutL.lex ().put (text ());
  push_state (yylloc, CCODE_BLOCK);
  nLbraces = 0;
}

"%init" {
  reflexDirective (); // RE/flex unique option
  yoxOutL.lex ().put (text ());
  push_state (yylloc, CCODE_BLOCK);
  nLbraces = 0;
}

"%reentrant" {
  reflexDirective (); // RE/flex unique option = Flex "%option reentrant"
}

"%top" {
  yoxOutL.lex ().put (text ());
  push_state (yylloc, CCODE_BLOCK);
  nLbraces = 0;
}

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

^{OPTWS}"%%".*$ {
  yylloc.begin.column += std::strcspn (text (), "%");
  yylloc.end = yylloc.begin + 2;
  separatorMARK.put (text ());
  start (RULES_PROLOG);
  return (lexer::token::MARK);
}

"%o"("ption")?{WS}.*$ {
  yoxOutL.lex ().put (text ());
  // check the '%option' list for any of interest to Ox
  // '%o' is a short-form (RE/flex)
  const std::string& lexeme (text ());
  const int optLeng = (int)strlen ("%option");
  if (lexeme.substr (0, optLeng) == "%option")
    {
      // Flex or RE/flex '%option' directive
      checkLexGenOptionDirective (yylloc, lexeme.substr (optLeng));
    }
  else
    {
      try
        {
          // Lex 'output slots' option, if '%o' followed by an integer
          std::stoi (lexeme.substr (2));
        }
      catch (std::invalid_argument const&)
        {
          // RE/flex unique short form of '%option'
          reflexDirective ();
          checkLexGenOptionDirective (yylloc, lexeme.substr (2));
        }
    }
}

"//".*$ {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

<<EOF>> {
  // EOF w/o finding the first "%%": the L-file is corrupt, so just quit
  unexpectedEOF (yylloc);
}

}



<LEX_CCODE>{

  /* LEX_DEFS, LEX_REGEXP 'subroutine': echo C/C++ code between "%{" and "%}" */

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

^"%}".*$ {
  yoxOutL.lex ().put (text ());
  pop_state ();
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<CCODE_BLOCK>{

  /* LEX_DEFS 'subroutine': echo C/C++ code "{ ... }" for Flex extension "%top"
     and for RE/flex extensions "%class" and "%init" */

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

"{" {
  yoxOutL.lex ().put ((char)chr ());
  nLbraces++;
}

"}" {
  yoxOutL.lex ().put ((char)chr ());
  if (--nLbraces == 0)
    pop_state (); // -> LEX_DEFS
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<RULES_PROLOG>{

  /* echo the prolog of the Lex Rules Section */

^{NON_WS} {
  yyless (0);
  start (LEX_REGEXP);
  nLcurls = 0;
  nLbraces = 0;
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

}



<LEX_REGEXP>{

  /* echo Lex regular expressions, terminated by whitespace */

^{OPTWS}"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

^{OPTWS}"%%".*$ {
  yylloc.begin.column += std::strcspn (text (), "%");
  yylloc.end = yylloc.begin + 2;
  yoxOutL.lex ().put (text ());
  start (LEX_USER);
  return (lexer::token::MARK);
}

^"%{".*$ {
  yoxOutL.lex ().put (text ());
  push_state (yylloc, LEX_CCODE);
}

{WS} {
  yoxOutL.lex ().put (text ());
  push_state (yylloc.end, LEX_ACT); // starting position if EOF in lexer action
}

^{OPTWS}"//".*$ {
  yoxOutL.lex ().put (text ());
}

{WS}"|".*$ {
  yoxOutL.lex ().put (text ());
}

{WS}$ {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

}



<LEX_ACT>{

  /* look for the "return" reserved word in a Lex action (-> LEX_RTN_OBJS), and
     look for the end of a Lex action after an attribute reference section; a
     Lex action is terminated by the first newline outside of paired braces */

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

"%{" {
  yoxOutL.lex ().put (text ());
  ++nLcurls;
  return (lexer::token::LCURL);
}

"%}" {
  yoxOutL.lex ().put (text ());
  --nLcurls;
  return (lexer::token::RCURL);
}

"{" {
  yoxOutL.lex ().put ((char)chr ());
  ++nLbraces;
  return (lexer::token::LBRACE);
}

"}" {
  yoxOutL.lex ().put ((char)chr ());
  --nLbraces;
  return (lexer::token::RBRACE);
}

"return" {
  returnSTMT.lex ().put (text ());
  nLparens = 0;
  start (LEX_RTN_OBJS);
  return (lexer::token::RETURN);
}

\n {
  yoxOutL.lex ().put ((char)chr ());
  if (nLbraces == 0 && nLcurls == 0)
    {
      pop_state (); // -> LEX_REGEXP, use the EOL position if syntax error
      yylloc.end = yylloc.begin;
      return (lexer::token::NEWLINE);
    }
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

. {
  yoxOutL.put ((char)chr ());
}

<<EOF>> {
  unterminatedAction (yylloc, startPos.top ());
}

}



<GEN_ACT>{

  /* look for the "return" reserved word in C/C++ source (-> GEN_RTN_OBJS) */

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

"{" {
  yoxOutL.lex ().put ((char)chr ());
  return (lexer::token::LBRACE);
}

"}" {
  yoxOutL.lex ().put ((char)chr ());
  return (lexer::token::RBRACE);
}

"return" {
  returnSTMT.lex ().put (text ());
  nLparens = 0;
  start (GEN_RTN_OBJS);
  return (lexer::token::RETURN);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

}



<LEX_RTN_OBJS>{

  /* token(s) (name, character literal) following the "return" in a Lex action
     up to the terminating ';' are put in a lookahead buffer, until the
     attribute reference section (if any) is processed */

\n {
  // return action statement broken by line feed

  returnSTMT.lex ().put ((char)chr ());
  if (nLbraces == 0 && nLcurls == 0)
    yylloc.end = yylloc.begin;
  return (lexer::token::NEWLINE); // force a syntax error @ the EOL position
}

";" {
  returnSTMT.lex ().put ((char)chr ());
  start (LEX_IS_ATTR_REF);
  return (lexer::token::SEMICOLON);
}

}



<GEN_RTN_OBJS>{

  /* token(s) (name, character literal) following the "return" up to the
     terminating ';' are put in a lookahead buffer, until the attribute
     reference section (if any) is processed */

\n {
  returnSTMT.lex ().put ((char)chr ());
}

";" {
  returnSTMT.lex ().put ((char)chr ());
  start (GEN_IS_ATTR_REF);
  return (lexer::token::SEMICOLON);
}

}



<LEX_RTN_OBJS,GEN_RTN_OBJS>{

"/*" {
  returnSTMT.lex ().put (text ());
  yox_push_state (COMMENT, returnSTMT);
}

"L"?"\"" {
  returnSTMT.lex ().put (text ());
  yox_push_state (STRING, returnSTMT);
}

"(" {
  returnSTMT.lex ().put ((char)chr ());
  nLparens++;
}

")" {
  returnSTMT.lex ().put ((char)chr ());
  if (--nLparens < 0)
    unmatchedRPAREN (yylloc);
}

{CHAR_CONST} {
  returnSTMT.lex ().put (text ());

  const std::string& lexeme (text ());
  const std::string& Name (stringLookup (lexeme, false));

  if (!Name.empty ())
    {
      if (!gSymLookup (Name).empty ())
        {
          if (retObjectSeen && atLeastOneSymHasAttrbs && !multipleReturnsIssued)
            {
              multipleReturnTokens (yylloc);
              multipleReturnsIssued = true;
            }
          retObjectSeen = true;
          yylval.emplace<char_literal> (Name);
          return (lexer::token::TERM_LITERAL);
        }
      else
        {
          yylval.emplace<char_literal> (Name);
          return (lexer::token::LITERAL_CHAR);
        }
    }
  else
    {
      yylval.emplace<char_literal> (stringLookup (lexeme, true));
      return (lexer::token::LITERAL_CHAR);
    }
}

{ID} {
  returnSTMT.lex ().put (text ());

  // skip the "api.token.prefix" if found, since the
  // token does not have the prefix in the gSymTable
  std::string lexeme (text ());
  lexeme = stripTokenPrefix (lexeme);

  const std::string& Name (stringLookup (lexeme, false));

  if (!Name.empty ())
    {
      if (!gSymLookup (Name).empty ())
        {
          if (gSymTable[Name].symNum < 0)
            returnNonterminalAsToken (yylloc, Name);
          if (retObjectSeen && atLeastOneSymHasAttrbs && !multipleReturnsIssued)
            {
              multipleReturnTokens (yylloc);
              multipleReturnsIssued = true;
            }
          retObjectSeen = true;
          yylval.emplace<name_literal> (Name);
          return (lexer::token::TERM_ID);
        }
      else
        {
          yylval.emplace<name_literal> (Name);
          return (lexer::token::ID);
        }
    }
  else
    {
      yylval.emplace<name_literal> (stringLookup (lexeme, true));
      return (lexer::token::ID);
    }
}

{DOT_NAME} {
  // not a legal terminal symbol: either a nonterminal or undefined

  returnSTMT.lex ().put (text ());

  // skip the "api.token.prefix" if found, since the token does not have the
  // prefix in the gSymTable
  std::string lexeme (text ());
  lexeme = stripTokenPrefix (lexeme);

  const std::string& Name (stringLookup (lexeme, false));

  if (!Name.empty ())
    {
      if (!gSymLookup (Name).empty ())
        {
          if (gSymTable[Name].symNum < 0)
            returnNonterminalAsToken (yylloc, Name);
          if (retObjectSeen && atLeastOneSymHasAttrbs && !multipleReturnsIssued)
            {
              multipleReturnTokens (yylloc);
              multipleReturnsIssued = true;
            }
          retObjectSeen = true;
          yylval.emplace<name_literal> (Name);
          return (lexer::token::TERM_ID);
        }
      else
        {
          yylval.emplace<name_literal> (Name);
          return (lexer::token::ID);
        }
    }
  else
    {
      yylval.emplace<name_literal> (stringLookup (lexeme, true));
      return (lexer::token::ID);
    }
}

"//".*$ {
  returnSTMT.lex ().put (text ());
}

. {
  returnSTMT.put ((char)chr ());
}

}



<LEX_IS_ATTR_REF>{

  /* look for a "@{" starting an attribute reference section; a '\n' outside
     of paired braces, or anything else except whitespace and comments, means
     that there is no attribute reference section */

{WS}"@{" {
  // throw away whitespace preceding "@{" "}"<-balance lbrace
  yyless (size () - 2); // rescan "@{"
}

{WS}$ {
  // throw away whitespace preceding '\n'
  ; // no-op
}

"/*" {
  returnSTMTtrailer.lex ().put (text ());
  yox_push_state (COMMENT, returnSTMTtrailer);
}

"@{" {
  push_state (yylloc, LEX_ATTR_REF);
  return (lexer::token::OX_LBRACE);
}

"//".*$|{WS} {
  returnSTMTtrailer.lex ().put (text ());
}

\n {
  returnSTMTtrailer.lex ().put ((char)chr ());
  if (nLbraces == 0 && nLcurls == 0)
    {
      pop_state (); // -> LEX_REGEXP, use the EOL position if syntax error
      yylloc.end = yylloc.begin;
      return (lexer::token::NEWLINE);
    }
}

. {
  yyless (0);
  start (LEX_ACT);
  // hack: hook for a Yacc action at end of a 'return'
  // statement with no attribute reference section
  return (lexer::token::NO_ATTRBREFSEC);
}

}



<GEN_IS_ATTR_REF>{

  /* look for a "@{" starting an attribute reference section; anything other
     than '\n', "@{", whitespace or comments means that there is no attribute
     reference section */

{WS}"@{" {
  // throw away whitespace preceding "@{" "}"<-balance lbrace
  yyless (size () - 2); // rescan "@{"
}

{WS}$ {
  // throw away whitespace preceding '\n'
  ; // no-op
}

"/*" {
  returnSTMTtrailer.lex ().put (text ());
  yox_push_state (COMMENT, returnSTMTtrailer);
}

"@{" {
  push_state (yylloc, GEN_ATTR_REF);
  return (lexer::token::OX_LBRACE);
}

"//".*$|{WS}|\n {
  returnSTMTtrailer.lex ().put ((char)chr ());
}

. {
  yyless (0);
  start (GEN_ACT);
  // hack: hook for a Yacc action at end of a 'return'
  // statement with no attribute reference section
  return (lexer::token::NO_ATTRBREFSEC);
}

}



<LEX_ATTR_REF>{

  /* Ox L-file attribute reference section */

"@}" {
  pop_state (); // -> LEX_IS_ATTR_REF
  start (LEX_ACT);
  return (lexer::token::OX_RBRACE);
}

}



<GEN_ATTR_REF>{

  /* Ox Generic L-file attribute reference section */

"@}" {
  pop_state (); // -> GEN_IS_ATTR_REF
  start (GEN_ACT);
  return (lexer::token::OX_RBRACE);
}

}



<LEX_ATTR_REF,GEN_ATTR_REF>{

{CPP_DIRECTIVE} {
  yoxOutL.lex ().put (text ());

  if (text ()[(int)size () - 1] == '\\') // multi-line directive
    {
      // pass thru the following '\n'
      yoxOutL.lex ().put ((char)yyinput ());
      yox_push_state (CPP_DIRECTIVE, yoxOutL);
    }
}

{WS}"@}" {
  // throw away whitespace preceding "@}"
  yyless (size () - 2); // rescan "@}"
}

{WS}$ {
  // throw away whitespace preceding '\n'
  ; // no-op
}

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

{ID} {
  const std::string& lexeme (text ());
  const std::string& Name (stringLookup (lexeme, true));

  do
    {
      if (expandingMacro ())
        {
          // check to see if 'Name' is a macro formal parameter name
          int argIndex = getCurMacro ().macroFormalLookup (Name);
          if (argIndex != -1)
            {
              std::string substValue (substFormal (argIndex));
              position arg_origin (yylloc.begin);

              // setup new input buffer for macro argument
              Matcher* newbuf =
                  new_matcher (new std::istringstream (substValue));
              pushbuffer (newbuf, arg_origin); // make it current
              setupFormalExpansion (yylloc);
              break;
            }
        }

      if (macroLookup (Name))
        {
          nLparensM = 0;
          push_state (yylloc, EXPAND_MACRO);
          expandMacro (Name, yylloc.begin);
        }
      else
        yoxOutL.lex ().put (Name);
    }
  while (false);
}

"@{" {
  unexpectedATLBRACE (yylloc);
}

{ATTRB_REF} {
  int nAt = 1; // # of '@'
  if (text ()[(int)size () - 1] == '@')
    ++nAt;
  else
    yox_warning (yylloc, "missing final '@': \"" + str () + "\".\n");

  std::string attrb_occur (text (), 1, size () - nAt);

  if (expandingMacro ())
    expandFormalsInAttrOccr (attrb_occur);

  yylval.emplace<attrbTriple> (scanAttrbName (yylloc, attrb_occur));
  return (lexer::token::OX_ATTRB_REF);
}

";" {
  yoxOutL.lex ().put ((char)chr ());
  return (lexer::token::SEMICOLON);
}

"@" {
  unexpectedChar (yylloc, '@', "in attribute reference section");
}

"//".*$|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.lex ().put ((char)chr ());
}

<<EOF>> {
  if (replacingFormal ())
    {
      // current macro argument complete
      popbuffer; // revert to previous input buffer
      macroFormalDone ();
    }
  else if (expandingMacro ())
    {
      // current macro expansion complete
      popbuffer; // revert to previous input buffer
      macroDone ();
    }
  else
    unterminatedAttrRefSection (yylloc, startPos.top ());
}

}



<EXPAND_MACRO>{

  /* LEX_ATTR_REF, GEN_ATTR_REF 'subroutine': expand a macro */

"/*" {
  yoxOutL.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.lex ().put (text ());
  yox_push_state (STRING, yoxOutL);
}

"(" {
  if (++nLparensM == 1)
    {
      // start of a new argument
      if (theFiles.scanner == LEX_SPEC_LEXER)
        macroARG = macroArg (yylloc.end, true); // -> Lex lexer output
      else
        macroARG = macroArg (yylloc.end, false); // -> Gen lexer output
      push_state (yylloc, MACRO_ARG);
    }
  else
    unbalancedParensInMacro (yylloc);
}

")" {
  if (nLparensM == 0)
    {
      macroARG.terminate ();

      std::string macroBody;
      position body_origin;
      setupMacroExpansion (yylloc, macroBody, body_origin);

      // setup new input buffer for macro expansion
      Matcher* newbuf =
          new_matcher (new std::istringstream (macroBody));
      pushbuffer (newbuf, body_origin); // make it current
      pop_state ();
    }
  else
    unbalancedParensInMacro (yylloc);
}

"," {
  macroARG.terminate ();
  // start of a new argument
  if (theFiles.scanner == LEX_SPEC_LEXER)
    macroARG = macroArg (yylloc.end, true); // -> Lex lexer output
  else
    macroARG = macroArg (yylloc.end, false); // -> Gen lexer output
  push_state (yylloc, MACRO_ARG);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.lex ().put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<MACRO_ARG>{

  /* EXPAND_MACRO 'subroutine': process a macro argument */

"/*" {
  macroARG.put (text ());
  yox_push_state (COMMENT, macroARG);
}

"L"?"\"" {
  macroARG.put (text ());
  yox_push_state (STRING, macroARG);
}

"(" {
  if (++nLparensM > 1)
    {
      // nested expression in the argument
      macroARG.put ((char)chr ());
    }
  else
    unbalancedParensInMacro (yylloc);
}

")" {
  if (--nLparensM > 0)
    macroARG.put ((char)chr ());
  else if (nLparensM == 0)
    {
      yyless (0);
      pop_state (); // -> EXPAND_MACRO
    }
  else
    unbalancedParensInMacro (yylloc);
}

"," {
  if (nLparensM > 1)
    macroARG.put ((char)chr ());
  else if (nLparensM == 1)
    {
      yyless (0);
      pop_state (); // -> EXPAND_MACRO
    }
  else
    unbalancedParensInMacro (yylloc);
}

\\\\ {
  macroARG.put ('\\');
}

\\, {
  macroARG.put (',');
}

\\\( {
  macroARG.put ('(');
}

\\\) {
  macroARG.put (')');
}

\\ {
  escapeCharMisuse (yylloc);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  macroARG.put (text ());
}

\n|. {
  macroARG.put ((char)chr ());
}

}



<LEX_USER>{

  /* start of the optional user section following the second "%%" in the L-file,
     if present */

\n|. {
  yoxOutL.lex ().put ((char)chr ());
  start (LEX_TRAILER);
}

}



<LEX_TRAILER>{

  /* echo the optional user section */

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.put (text ());
  yox_push_state (STRING, yoxOutL);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<COMMENT>{

  /* 'subroutine': handle multi-line C comments */

"*/" {
  whereTo->put (text ());
  pop_state ();
}

[^\*\n]+ {
  whereTo->put (text ());
}

\n|. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedComment (yylloc, startPos.top ());
}

}



<STRING>{

  /* 'subroutine': handle multi-line C/C++ string literals */

"\"" {
  whereTo->put ((char)chr ());
  if (stringHasBareNewline)
    stringContainsNewline (yylloc, startPos.top ());
  pop_state ();
}

[^"\\\n]+|\\(.|\n) {
  whereTo->put (text ());
}

\n {
  stringHasBareNewline = true;
  whereTo->put ((char)chr ());
}

. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedString (yylloc, startPos.top ());
}

}



<CPP_DIRECTIVE>{

  /* 'subroutine': handle C/C++ directives */

\\\n {
  // C/C++ directive spans multiple source lines
  whereTo->put (text ());
}

\n {
  whereTo->put ((char)chr ());
  pop_state ();
}

. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedCPPdirective (yylloc, startPos.top ());
}

}



<<EOF>> {
  return lexer::token::END;
}

%%

void
REFLEX_OPTION_namespace::REFLEX_OPTION_lexer::yox_push_state (int lexState,
                                                              outMgr& whTo)
{
  whereTo = &whTo;
  if (lexState == STRING)
    stringHasBareNewline = false;
  push_state (*whereTo->getCurPos (), lexState);
}



// The main problem is for this lexer to know when it is looking at a Lex
// action, to look for '@'s inside such actions, and to ignore '@'s elsewhere.

// Lexemes suitable for grammar symbols are 1) identifiers previously declared
// as tokens, or 2) character constants.

// If exactly one lexeme suitable for a grammar symbol appears between return
// and ';', a node of that type will be put on the tree just before each time
// said return is executed.

// If no such lexeme appears there, a warning is issued.  If there is not a
// non-empty attribute definition section, the type of the dummy node symbol is
// used and another warning is issued.

// If more than one suitable lexeme appears, the first such is assumed for the
// node type, and subsequent ones generate a warning.
