/* $Id: scan.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2007-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include "location.hh"

#include "fileMgr.hh"
#include "lextypes.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "symTrans.hh"


// Y-file processing:

Yfile_specIO Yfile;

fileMgr yoxOutGdefs;
fileMgr yoxOutG;
fileMgr yoxOutGimport;

fileMgr yoxOutH;
fileMgr yoxOutHrequires;
fileMgr yoxOutHprovides;

// hold the closing '}' of a Yacc rule action
lookaheadBuf actionCLOSER;

// hold accumulated Ox attrinbute component declarations
lookaheadBuf attrDECL;

// hold potential rule LHS: (symbol, inherited attributes, ':')
lookaheadBuf possibleLHS;

// hold any rule directives immediately following the last Yacc action
lookaheadBuf ruleDIRECTIVE;

// hold the "%union" component declarations
lookaheadBuf unionBODY ("%union body", false);

// hold accumulated Yacc code blocks
lookaheadBuf yaccCODE;

refSegBuf rsBuffer; // the only refSegBuf


// L-file processing:

std::vector<class Lfile_specIO> Lfile;

fileMgr yoxOutL (true); // -> Lex lexer output

// hold the "return ... ;" from a lexer action
lookaheadBuf returnSTMT (true);

// hold whatever follows "return ... ;" up to the start of an
// attribute reference section, or the end of the lexer action
lookaheadBuf returnSTMTtrailer (true);


// used by the Y-file & L-file lexers:

lookaheadBuf separatorMARK; // hold a "%%" line



bool retObjectSeen = false;
bool multipleReturnsIssued = false;

// increment for each `%union' found in the grammar definition file;
// if > 0 ensure that Yacc code blocks occur after the new `%union'
int yaccUnionDefined = 0;

// increment for each unsupported `api.value.type' definition found
int bisonApiValueTypeUnsupported = 0;

// =true if `api.value.type' defined as `union' in the parser definition file
bool bisonApiValueTypeUnionDefined = false;



// return true if Yacc `%union' and/or Bison "%define api.value.type"
// declarations have occurred
bool
unionDefined (void)
{
  return (yaccUnionDefined != 0 ||
          (outMgr::ctrl.targetParserGen == bisonParserGen &&
           bisonApiValueTypeUnionDefined));
}



static std::string
checkIfGsym (const std::string token)
{
  const int len = (int)token.length ();
  std::string local;

  if (((token[0] == '\'') && (token[len - 1] == '\'')) ||
      ((token[0] == '\"') && (token[len - 1] == '\"')))
    local = stringLookup (token, true);
  else
    // skip the "api.token.prefix" if found, since the
    // token doesn't have the prefix in the gSymTable
    local = stringLookup (stripTokenPrefix (token), false);

  if (local.empty ())
    return local;

  // if token is an alias, local is the aliased symbol
  return gSymLookup (local);
}



void
getAttrRefComponents (const std::string& token,
                      std::string& gsymName,
                      std::string& occurrence,
                      std::string& attrName)
{
  // there will always be at least one '.'
  const int last_dot = (int)token.find_last_of ('.');
  attrName = token.substr (last_dot + 1);

  // there may be a '.' followed by an occurrence integer preceding the last '.'
  // use 'find_last_of ()' in case the grammar symbol contains '.'
  const int first_dot = (int)token.substr (0, last_dot).find_last_of ('.');
  if (first_dot != (long)std::string::npos)
    {
      occurrence = token.substr (first_dot + 1, last_dot - first_dot - 1);

      // if what precedes the next-to-last '.' is a grammar symbol, we are done
      gsymName = token.substr (0, first_dot);
    }
  else
    {
      gsymName = token.substr (0, last_dot);
    }
}



// trim off leading & trailing whitespace
const std::string
trim (std::string& token)
{
  const std::string WS = " \t\f\v";
  int start = token.find_first_not_of (WS),
      end = token.find_last_not_of (WS);
  return token.substr (start, end - start + 1);
}



attrbTriple
scanAttrbName (const AGC::location& refLoc, const std::string& src_token)
{
  attrbTriple result;
  result.occurrence = 0;

  std::string gsymName, occurrence, attrName;
  getAttrRefComponents (src_token, gsymName, occurrence, attrName);

  result.gsName = checkIfGsym (trim (gsymName));
  if (!occurrence.empty ())
    result.occurrence = std::stoi (occurrence, nullptr, 10);
  result.atName = stringLookup (trim (attrName), false);

  if (result.atName.empty () || result.gsName.empty ())
    {
      yox_error (refLoc,
                 "unknown attribute occurrence: \"" + src_token + "\".\n");
      result.occurrence = -1;
    }

  return result;
}



static const std::regex
    lineDirRegexp ("^ *# *line +([[:digit:]]+) +\"([^\"]*)\"",
                   std::regex::extended);

long
getLineDir (const char* buffer, int& lineno)
{
  std::cmatch m;

  if (std::regex_search (buffer, m, lineDirRegexp))
    {
      // #line directive found
      lineno = std::stoi (std::string (buffer).substr (m.position (1)));
      return m.position (0) + m.length (0) - 1;
    }
  else
    return 0;
}

long
getLineDir (const char* buffer, std::string& fileName)
{
  std::cmatch m;

  if (std::regex_search (buffer, m, lineDirRegexp))
    {
      // #line directive found
      fileName = std::string (buffer).substr (m.position (2), m.length (2));
      return m.position (0) + m.length (0) - 1;
    }
  else
    return 0;
}
