/* $Id: scan.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2007-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef SCAN_HH_INCLUDED
#define SCAN_HH_INCLUDED

#include <regex>
#include <vector>

#include "location.hh"

#include "fileMgr.hh"
#include "lextypes.hh"
#include "memMgr.hh"
#include "opts.hh"


// track EOL/EOF context of most recent lexer match, for location calculations
class EOLctx
{
public:

  bool atEOL;           // => true if at EOL
  bool atEND;           // => true if at EOF
  AGC::position EOLpos; // EOL position, if 'atEOL'

  EOLctx (AGC::position pos)
      : atEOL (false),
        atEND (false),
        EOLpos (pos)
  {}

  EOLctx (bool EOL, bool END, AGC::position pos)
      : atEOL (EOL),
        atEND (END),
        EOLpos (pos)
  {}
};



// use 'scanner' for translation of file 'inName' -> 'outName'
class specIO
{
public:

  enum whichYoxLexer scanner;

  std::string inName;
  std::string outName;

  specIO (enum whichYoxLexer which)
      : scanner (which),
        inName (""),
        outName ("")
  {}

  specIO (std::string in, enum whichYoxLexer which, std::string out)
      : scanner (which),
        inName (in),
        outName (out)
  {}
};

class Yfile_specIO : public specIO
{
public:

  Yfile_specIO ()
      : specIO (YACC_SPEC_LEXER)
  {}

  Yfile_specIO (std::string in, std::string out)
      : specIO (in, YACC_SPEC_LEXER, out)
  {}
};

class Lfile_specIO : public specIO
{
public:

  Lfile_specIO ()
      : specIO (NO_LEXER)
  {}

  Lfile_specIO (std::string in, enum whichYoxLexer which, std::string out)
      : specIO (in, which, out)
  {}
};



// Y-file processing:

extern Yfile_specIO Yfile;

extern fileMgr yoxOutGdefs; // Y-file definition section code
extern fileMgr yoxOutG;
extern fileMgr yoxOutGimport; // holds "%import" `msta' extension

extern fileMgr yoxOutH;
extern fileMgr yoxOutHrequires; // if can generate "%code requires" for Bison
extern fileMgr yoxOutHprovides; // if can generate "%code provides" for Bison

// hold the closing '}' of a Yacc rule action
extern lookaheadBuf actionCLOSER;

// hold accumulated Ox attribute component declarations
extern lookaheadBuf attrDECL;

// hold potential rule LHS: (symbol, inherited attributes, ':')
extern lookaheadBuf possibleLHS;

// hold any rule directives immediately following the last Yacc action
extern lookaheadBuf ruleDIRECTIVE;

// hold the "%union" component declarations
extern lookaheadBuf unionBODY;

// hold accumulated Yacc code blocks
extern lookaheadBuf yaccCODE;

// delimiter character used for attribute number pairs in strings
#define attrbNumPairDelimChar '\037'

extern refSegBuf rsBuffer;



// L-file processing:

extern fileMgr yoxOutL;

extern std::vector<class Lfile_specIO> Lfile;

// hold the "return ... ;" from a lexer action
extern lookaheadBuf returnSTMT;

// hold whatever follows "return ... ;" up to the start of an
// attribute reference section, or the end of the lexer action
extern lookaheadBuf returnSTMTtrailer;

extern bool retObjectSeen;
extern bool multipleReturnsIssued;



// Y-file & L-file processing:

extern lookaheadBuf separatorMARK; // hold a "%%" line



extern int yaccUnionDefined;

extern int bisonApiValueTypeUnsupported;

// =true if `api.value.type' defined as `union' in the parser definition file
extern bool bisonApiValueTypeUnionDefined;

// return true if Yacc `%union' and/or Bison "%define api.value.type"
// declarations have occurred
extern bool unionDefined (void);


extern void getAttrRefComponents (const std::string&,
                                  std::string&,
                                  std::string&,
                                  std::string&);


// trim off leading & trailing whitespace
extern const std::string trim (std::string&);


extern attrbTriple scanAttrbName (const AGC::location&, const std::string&);


extern long getLineDir (const char*, int&);
extern long getLineDir (const char*, std::string&);

#endif // !SCAN_HH_INCLUDED
