/* $Id: symTrans.cc 2839 2022-12-13 00:53:33Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <iomanip>
#include <sstream>

#include <unordered_map>

#include "location.hh"

#include "depGs.hh"
#include "fileMgr.hh"
#include "joinTabs.hh"
#include "lextypes.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "semtypes.hh"
#include "symTrans.hh"
#include "treeviz.hh"



// Here follow some data structures for giving serial numbers to objects
// encountered in Ox's input.  These sequences most of the time hold the
// value of the *next* serial number to be used, rather than the one that
// was *just* used.

//******************************************************************************


#define firstTermNum 1

// increasing sequence of numbers to represent terminals of the grammar
static long termSeq = firstTermNum;



#define firstNonTermNum -1

// decreasing sequence of numbers to represent nonterminals of the grammar
static int nontermSeq = firstNonTermNum;



static long startSymNum = 0;

std::string startSymLexeme;



// increasing sequence of numbers to represent productions of the grammar
#define firstProdNum 1
int prodNumSeq = firstProdNum;



// increasing sequence of numbers representing each declared attribute;
// corresponds to the index in yyySorts[] in the generated evaluator
int attrbDeclAttrbOccurSeq = 0;



// =true if at least one grammar symbol appears in some attribute declaration
bool atLeastOneSymHasAttrbs = false;


//******************************************************************************


// expansion increment for the index to dependents by production number
#define PRODSINCR 1024

std::vector<long> indToEntsByProdNum (PRODSINCR + 1); // starts w/ production #1


//******************************************************************************


// common 'not found' empty string
static const std::string notFound = std::string ();



// The 'stringTable []' contains copies of lexemes (other than
// those for Ox reserved words) found in Ox's input.
// These strings will mostly be:
//   1) identifiers, and
//   2) strings that represent character constants.

static int nextStringSerialNo = 0; // next string serial number


typedef std::unordered_map<std::string, int> stringTableType;

static stringTableType stringTable;



// the table of grammar terminal symbol aliases
struct gSymAlias
{
  std::string token;

  gSymAlias ()
      : token (std::string ())
  {}
};

typedef std::unordered_map<std::string, gSymAlias> gSymAliasTableType;

static gSymAliasTableType gSymAliasTable;


static void
addTo_gSymAliasTable (const AGC::location& loc, const std::string& alias)
{
  try
    {
      (void)gSymAliasTable.at (alias);
      yox_error (loc,
                 "token alias \"" + alias + "\" is already defined for token "
                 "\"" + gSymAliasTable[alias].token + "\".\n");
    }
  catch (std::out_of_range const&)
    {
      try
        {
          // not present, add it
          (void)gSymAliasTable.insert (
              gSymAliasTableType::value_type (alias, gSymAlias ()));
        }
      catch (...)
        {
          // insert on the alias table failed
          yox_alloc_fail (__FILE__, __LINE__);
        }
    }
}

// get the base grammar token, given its string literal alias (if it exists)
static const std::string&
gSymAliasLookup (const std::string& alias)
{
  try
    {
      return gSymAliasTable.at (alias).token;
    }
  catch (std::out_of_range const&)
    {
      return (notFound);
    }
}



// table of grammar symbols & associated attribute definitions
gSymTableType gSymTable;



static void
addTo_gSymTableIfNew (const std::string& symbol)
{
  try
    {
      (void)gSymTable.at (symbol);
    }
  catch (std::out_of_range const&)
    {
      try
        {
          // not present, add it
          (void)gSymTable.insert (gSymTableType::value_type (symbol, gSym ()));
        }
      catch (...)
        {
          // insert on the symbol table failed
          yox_alloc_fail (__FILE__, __LINE__);
        }
    }
}

// lookup the symbol in the grammar symbol table;
// if not there, look in the token alias table
const std::string&
gSymLookup (const std::string& symbol)
{
  const yoxControl& ctrl = outMgr::ctrl;

  try
    {
      (void)gSymTable.at (symbol);
      return symbol;
    }
  catch (std::out_of_range const&)
    {
      if (ctrl.targetParserGen == bisonParserGen ||
          ctrl.targetParserGen == bisonParserGenOld)
        {
          // check to see if this is a token alias: either return the
          // base token, or notFound
          return gSymAliasLookup (symbol);
        }
      else
        return (notFound);
    }
}



struct attrbSym
{
  attrbAnnotation autodef; // copy rule annotation associated with the attribute

  attrbSym ()
      : autodef (attrbAnnotation ())
  {}
};

typedef std::unordered_map<std::string, attrbSym> attrbSymTableType;

// the table of attributes
// - attribute names are globally unique
// - attributes of the same name in multiple sorts must have the same copy rule
//   direction annotation (checked by Ox)
// - the copy rule '@warn' annotation is cumulative, if not consistently applied
// - same named attributes in multiple sorts must be of the same C/C++ type
//   - this is not currently checked by Ox, so mismatched type
//     errors will not be found until compile-time
static attrbSymTableType attrbSymTable;


static void
addTo_attrbSymTableIfNew (const std::string symbol)
{
  try
    {
      (void)attrbSymTable.at (symbol);
    }
  catch (std::out_of_range const&)
    {
      try
        {
          // not present, add it
          (void)attrbSymTable.insert (
              attrbSymTableType::value_type (symbol, attrbSym ()));
        }
      catch (...)
        {
          // insert on the table of attributes failed
          yox_alloc_fail (__FILE__, __LINE__);
        }
    }
}


// lookup the attribute symbol, given its string literal alias: 'true' => exists
static bool
attrbSymLookup (const std::string& symbol)
{
  try
    {
      (void)attrbSymTable.at (symbol);
      return (true);
    }
  catch (std::out_of_range const&)
    {
      return (false);
    }
}



// track the Yacc position of RHS symbols in the rule; required since embedded
// Yacc actions occupy a position that Ox doesn't know about; LHS is position 0
int yaccSemIndex = 0;

// points into prodBuffer2
int prodBufferPointer2 = 0;

// expansion increment for the array of symbols in a production
#define PRODLENINCR 256

// the current largest number of symbols in one production
int maxProdLength = PRODLENINCR;

// holds information about the symbols of the rule being parsed
production prodBuffer2 (PRODLENINCR);



// Search the string table for a matching string.
// - If it is absent and insertion is requested, insert it, returning
//   a pointer to a permanent null-terminated copy.
// - If absent and insertion not requested, return null pointer.
// - If already present, return pointer to the null-terminated copy.

// Fatal error upon attempt to insert into full table.

const std::string&
stringLookup (const std::string& cp, bool insertIfAbsent)
{
  try
    {
      (void)stringTable.at (cp);
      return cp;
    }
  catch (std::out_of_range const&)
    {
      if (insertIfAbsent)
        try
           {
             // not present, add it if requested
             (void)stringTable.insert (
                 stringTableType::value_type (cp, nextStringSerialNo++));
             return cp;
           }
         catch (...)
           {
             // insert on the string table failed
             yox_alloc_fail (__FILE__, __LINE__);
           }
      else
        return (notFound);
    }
#if defined(__GNUC__)
  __builtin_unreachable ();
#endif
}



void
printStringTab (void)
{
  std::string genStringTab[nextStringSerialNo];

  for (const std::pair<const std::string, int>& the_entry : stringTable)
    {
      if (attrbSymLookup (the_entry.first))
        genStringTab[the_entry.second] = the_entry.first;
      else
        genStringTab[the_entry.second] = gSymLookup (the_entry.first);
    }

  for (int i = 0; i < nextStringSerialNo; i++)
    {
      if ((i % 5) == 0)
        tempStringTabFile.put ('\n');

      std::string& the_entry = genStringTab[i];

      if (!the_entry.empty ())
        {
          if (attrbSymLookup (the_entry))
            tempStringTabFile.put ('"').put (the_entry).put ("\",");
          else
            {
              const std::string& str = gSymLookup (the_entry);
              if (!str.empty ())
                {
                  if (str[0] == '\"')
                    {
                      tempStringTabFile.put ('"');

                      for (const char& c : str)
                        {
                          if (c == '\"')
                            tempStringTabFile.put ('\\');

                          tempStringTabFile.put (c);
                        }

                      tempStringTabFile.put ("\",");
                    }
                  else
                    tempStringTabFile.put ('"').put (str).put ("\",");
                }
              else
                tempStringTabFile.put ("0,");
            }
        }
      else
        tempStringTabFile.put ("0,");
    }
}



// for debugging:
void
showStringTab (void)
{
  int i = 0;
  for (const std::pair<const std::string, int>& the_entry : stringTable)
    {
      std::cerr << std::setw (5) << i << "   " << std::setw (5)
                << the_entry.second        // stringTable [i].serialNo
                << ": " << the_entry.first // stringTable [i].entry
                << "\n";
      i++;
    }
}



// Search the string table for 'cp'.
// When found, return the string's serial number.
// Never call unless the string has already been inserted into the table.
int
stringSerialNoLookup (const std::string& cp)
{
  try
    {
      return stringTable.at (cp);
    }
  catch (std::out_of_range const&)
    {
      // not present
      yox_internal_fail (__FILE__, __LINE__, "string not found\n");
    }
#if defined(__GNUC__)
  __builtin_unreachable ();
#endif
}



void
attrbAutoDef (const AGC::location& loc,
              const std::string& cp,
              enum attrbDirection autodef,
              bool warn)
{
  const std::string& the_entry = stringLookup (cp, false); // the entry exists

  addTo_attrbSymTableIfNew (the_entry);
  attrbSym& attrb = attrbSymTable[the_entry];

  enum attrbDirection& autoDefDirection = attrb.autodef.direction;
  int& autoDefLineno = attrb.autodef.lineno;

  if (autodef == inherited && autoDefDirection == synthesized)
    {
      yox_error (loc,
                 "conflicting @autosyn (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");

      return;
    }

  if (autodef == synthesized && autoDefDirection == inherited)
    {
      yox_error (loc,
                 "conflicting @autoinh (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");

      return;
    }

  if (autodef == synthesized && autoDefDirection == synthesized)
    yox_warning (loc,
                 "duplicate global @autosyn (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");

  if (autodef == inherited && autoDefDirection == inherited)
    yox_warning (loc,
                 "duplicate global @autoinh (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");

  if (autoDefDirection == undefined && autodef != undefined)
    {
      autoDefDirection = autodef;
      autoDefLineno = loc.begin.line;
    }
  attrb.autodef.autoWarn |= warn;
}



// There is a list of sortAttrb, the attribute sort, and a parallel list of
// symAttrbDirection, recording whether each attribute in the sort is
// inherited or synthesized, associated with each grammar symbol.

// For the current grammar symbol, set:
//   the attribute sort
//   the initial inherited/synthesized state of each attribute
//     (if there was a corresponding copy rule annotation)
void
addAttrbsToGsym (const AGC::location& loc,
                 const std::string& symbol,
                 const sort& gSymSort)
{
  atLeastOneSymHasAttrbs = true;
  gSym& entry = gSymTable[symbol];
  if (entry.symSort.typeNum != NOTYPE)
    yox_error (loc,
               "multiple @attributes declarations for grammar symbol: "
               "\"" + symbol + "\".\n");
  else
    {
      entry.symSort = gSymSort; // set the attribute sort

      const sortAttrbSeq& sortSeq = *gSymSort.attrbSeq;

      // initial inherited/synthesized state of each attribute
      entry.attrbDirections = attrbDirectionSeq (sortSeq.size ());
      for (int i = 0; i < (int)sortSeq.size (); i++)
        entry.attrbDirections[i] = sortSeq[i].autoDef.direction;
    }
}



// warn if symbol ends in ".<integer>"; attribute references may be ambiguous
static void
warnIfAmbiguousGsym (const AGC::location& loc, const std::string& symbol)
{
  const int last_dot = (int)symbol.find_last_of ('.');

  if (last_dot == (long)std::string::npos)
    {
      // no '.' in symbol
      return;
    }

  if ((long)symbol.length () == last_dot + 1)
    {
      // '.' is last character in symbol
      return;
    }

  int i = last_dot;
  while (symbol[++i] != '\0')
    if (!isdigit (symbol[i]))
      {
        // doesn't end with ".<integer>"
        return;
      }

  yox_warning (loc,
               "grammar symbol "
               "\"" + symbol + "\" suffix \"" + symbol.substr (last_dot) + "\""
               " may cause attribute reference ambiguities.\n");
}



void
installGsymAsTokenIfNew (const AGC::location& loc, const std::string& token)
{
  addTo_gSymTableIfNew (token);

  if (gSymTable[token].symNum == 0)
    {
      gSymTable[token].symNum = termSeq++;
      warnIfAmbiguousGsym (loc, token);
    }
  else if (gSymTable[token].symNum == startSymNum)
    yox_error (loc, "start symbol \"" + token + "\" redefined as token.\n");
}



void
installAliasToToken (const AGC::location& loc_alias,
                     const std::string& alias,
                     const AGC::location& loc_token __attribute__ ((unused)),
                     const std::string& token)
{
#ifndef NDEBUG
  // only used by 'assert ()' below
  const std::string& local = gSymLookup (token);
#endif
  assert (!local.empty () && local == token); // isn't already an alias

  addTo_gSymAliasTable (loc_alias, alias);

  gSymTable[token].alias = alias;
  gSymAliasTable[alias].token = token;
}



void
installStartSym (const AGC::location& loc, const std::string& symbol)
{
  startSymLexeme = symbol;
  installGsymAsNontermIfNew (loc, symbol);
}



void
installGsymAsNontermIfNew (const AGC::location& loc, const std::string& symbol)
{
  addTo_gSymTableIfNew (symbol);

  if (gSymTable[symbol].symNum == 0)
    {
      gSymTable[symbol].symNum = nontermSeq--;
      warnIfAmbiguousGsym (loc, symbol);
    }
  else if (gSymTable[symbol].symNum > 0)
    {
      if (symbol == startSymLexeme)
        yox_error (loc, "token \"" + symbol + "\" redefined as start symbol.\n");
      else
        yox_error (loc, "token \"" + symbol + "\" redefined as nonterminal.\n");
    }

  if (symbol == startSymLexeme)
    startSymNum = gSymTable[symbol].symNum;
}



void
installGsymIfNew (const AGC::location& loc, std::string& symbol)
{
  const std::string& local = gSymLookup (symbol);

  if (local.empty ())
    {
      // symbol isn't a grammar symbol and isn't a token alias

      if (symbol[0] == '\'' || symbol[0] == '\"')
        {
          // literal tokens are not always defined by `%token', `%left', etc.
          installGsymAsTokenIfNew (loc, symbol);
        }
      else
        {
          // symbol is a nonterminal: non-literal terminal symbols have already
          // been defined by `%token', %left, etc.
          installGsymAsNontermIfNew (loc, symbol);
        }
    }
  else
    {
      // symbol might be an alias, so use the base token (if it exists)
      symbol = local;
    }
}



// create a new attribute w/ copy rule annotations (if any)
sortAttrb
newAttrbID (const AGC::location& loc,
            const std::string& cp,
            enum attrbDirection autodef,
            bool warn)
{
  const yoxControl& ctrl = outMgr::ctrl;

  if (cp == NIOsym)
    {
      // NIO pseudo attribute

      if (ctrl.genTreeViz)
        yox_error (loc, "\"" + NIOsym + "\" is a reserved attribute name.\n");
      else
        yox_warning (loc,
                     "\"" + NIOsym + "\" would be a reserved attribute name "
                     "(using the \"-t\" command-line option).\n");
    }

  const std::string& the_entry = stringLookup (cp, false); // the entry exists

  addTo_attrbSymTableIfNew (the_entry);
  attrbAnnotation& cur_autodef = attrbSymTable[the_entry].autodef;

  enum attrbDirection autoDefDirection = cur_autodef.direction;
  int autoDefLineno = cur_autodef.lineno;

  if (autodef == inherited && autoDefDirection == synthesized)
    {
      yox_error (loc,
                 "conflicting global @autosyn (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");
      autodef = undefined;
    }

  if (autodef == synthesized && autoDefDirection == inherited)
    {
      yox_error (loc,
                 "conflicting global @autoinh (line " +
                     std::to_string (autoDefLineno) + ")"
                 " for attribute \"" + cp + "\".\n");
      autodef = undefined;
    }

  if (autoDefDirection == undefined && autodef != undefined)
    {
      autoDefDirection = autodef;
      autoDefLineno = loc.begin.line;
    }

  return sortAttrb (the_entry,
                    attrbAnnotation (autoDefDirection,
                                     autoDefLineno,
                                     cur_autodef.autoWarn || warn));
}



// write the sort information into the run-time tables
void
putAttrbSort (const sort& gSymSort)
{
  if (((gSymSort.typeNum - firstSortTypeNum) % 10) == 0)
    tempSortIndFile.put ('\n');

  tempSortIndFile.put (attrbDeclAttrbOccurSeq, 3).put (',');

  const sortAttrbSeq& sortSeq = *gSymSort.attrbSeq;

  for (const sortAttrb& attrb : sortSeq)
    {
      if ((attrbDeclAttrbOccurSeq % 10) == 0)
        tempSortNumsFile.put ('\n');

      tempSortNumsFile.put (stringSerialNoLookup (attrb.id), 5).put (',');

      attrbDeclAttrbOccurSeq++;
    }
}



std::string omittedRetSym; // gSymTable omitted Lexer return token

#define maxNwyes 256

// install & memoize the `omitted Lexer return token'
void
setupOmittedRetSym (void)
{
  std::string wyes ("__y");

  while (!stringLookup (wyes, false).empty ())
    wyes.append (1, 'y');

  omittedRetSym = stringLookup (wyes, true);

  const AGC::location loc; // dummy location

  installGsymAsTokenIfNew (loc, omittedRetSym);
}



std::string errorSym; // gSymTable `error' token

// install & memoize the `error' token
void
setupErrorSym (void)
{
  errorSym = stringLookup ("error", true);

  const AGC::location loc; // dummy location

  installGsymAsTokenIfNew (loc, errorSym);
}



void
incProdNumSeq (void)
{
  const int size = (int)indToEntsByProdNum.size ();
  if ((++prodNumSeq) == size)
    try
      {
        indToEntsByProdNum.resize (size + PRODSINCR);
      }
    catch (std::bad_alloc const&)
      {
        // reallocation of index to dependents by prod num failed
        yox_alloc_fail (__FILE__, __LINE__);
      }
}



void
putSymbolInProdBuffer2 (const std::string& token)
{
  int symOccr = 0;

  const int size = (int)prodBuffer2.size ();
  if (prodBufferPointer2 == size)
    try
      {
        prodBuffer2.resize (size + PRODLENINCR);
        maxProdLength += PRODLENINCR;
      }
    catch (std::bad_alloc const&)
      {
        // reallocation of production buffer failed
        yox_alloc_fail (__FILE__, __LINE__);
      }

  for (int i = 0; i < prodBufferPointer2; i++)
    if (prodBuffer2[i].lexeme == token)
      symOccr++;

  prodSymbol& slot = prodBuffer2[prodBufferPointer2];
  slot.symOccr = symOccr;
  slot.lexeme = token;
  slot.yaccSemIndex = yaccSemIndex;

  prodBufferPointer2++;
  yaccSemIndex++;
}



// translate: attrbTriple (gsName, occurrence, atName)
//         to attrbPair   (symPos, attrbNum)
attrbPair
transAttrbTripleToAttrbPair (const AGC::location& loc,
                             const attrbTriple& lexRef)
{
  const yoxControl& ctrl = outMgr::ctrl;

  attrbPair result;

  result.pos = 0;
  result.num = 0;

  if (lexRef.occurrence != -1)
    {
      // occurrence is NOT invalid

      while (result.pos < prodBufferPointer2)
        {
          if (prodBuffer2[result.pos].lexeme == lexRef.gsName)
            if (lexRef.occurrence == prodBuffer2[result.pos].symOccr)
              break;
          result.pos++;
        }

      if (result.pos != prodBufferPointer2)
        {
          // occurrence is a valid position in the rule

          if (ctrl.genTreeViz && lexRef.atName == NIOsym)
            {
              // the NIO pseudo attribute
              result.num = NIOsymAttrbNum; // @<symbol>.<n>.<NIOsym>@
              return result;
            }

          gSym& entry = gSymTable[prodBuffer2[result.pos].lexeme];
          while (result.num < entry.symSort.nAttrbs)
            {
              if ((*(entry.symSort.attrbSeq))[result.num].id == lexRef.atName)
                return result;
              else
                result.num++;
            }
        }

      yox_error (loc,
                 "\"" + lexRef.gsName + "." +
                     std::to_string (lexRef.occurrence) + "." +
                     lexRef.atName + "\""
                 " is not valid for this production.\n");
    }

  result.pos = -1;
  result.num = -1;
  return result;
}
