/* Infix notation calculator--calc */
%code requires /* -> header file, preceding the definition of YYSTYPE */
{
#if defined (__cplusplus)
#include <cstdbool>
#else
#include <stdbool.h>
#endif

#ifndef YY_TYPEDEF_YY_SCANNER_T
# define YY_TYPEDEF_YY_SCANNER_T
typedef void* yyscan_t;
#endif

  typedef struct
  {
    bool verbose; /* print the intermediate results */
    int value;    /* value of the last computation */
    int nerrs;    /* count of errors */
  } result;

#include "types.h"
}

/* scanner and error count are exchanged between main, yyparse and yylex */
%param {yyscan_t scanner}{result *res}

%code provides /* -> header file, following the definition of YYSTYPE */
{
  /* the expected prototype of yylex:
   *   1st argument must be named `yylval_param'
   *   2nd argument must be named `yyscanner'
   *   3rd argument must be named `res'
   */
#define YY_DECL \
  enum yytokentype yylex ( \
	YYSTYPE* yylval_param, \
	yyscan_t yyscanner, result *res)
  YY_DECL;

  void yyerror (yyscan_t, result *, const char *, ...);
}

%code
{
static int power (int, int);

static char node_label[100];
}

%define api.pure full
%define api.token.prefix {TOK_}
%define api.value.type union
%define parse.error verbose
%define parse.trace
%verbose

%token
  EQ     "="
  PLUS   "+"
  MINUS  "-"
  STAR   "*"
  SLASH  "/"
  UP     "^"
  LPAREN "("
  RPAREN ")"
  EOL    "end-of-line"
  EOF 0  "end-of-file"
;

%token <int> NUM "number"
%type <int> exp
%printer { fprintf (yyo, "%d", $$); } <int>

/* precedence (lowest to highest) and associativity */
%nonassoc "="		/* comparison */
%left "+" "-"
%left "*" "/"
%precedence UNARY	/* unary plus or minus */
%right "^"		/* exponentiation */

@attributes {semantic_value iVal;} exp "number"
@attributes {@void print;} line

@traversal PV

%%
input	:	line
	|	input line
	;

line	:	exp EOL			@{
					  @decorate
					  @e line.print: exp.iVal;
					    {res->value = @exp.iVal@;
					     if (res->verbose)
						fprintf (stderr, "%d\n", @exp.iVal@);
					    }
					@}
	|	error EOL		{ yyyabort(); yyerrok; }
					@{ @m line.print; {} @}
//	|	EOL			@{ @m @line.print@ {} @} // TES: poor syntax errors (3) for this missing ';'
	|	EOL			@{ @m @line.print@; {} @}
	;

exp	:	"number"		@{ @i @exp.iVal@ = @"number".iVal@;
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.iVal@);
					      agset (@exp.NIO@, "label", node_label);
					     }
					@}
	|	exp "=" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ == @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	exp "+" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ + @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	exp "-" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ - @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	exp "*" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ * @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	exp "/" exp		{
					 if ($3 == 0)
					   {yyerror (scanner, res, "invalid division by zero");
					    YYERROR;
					   }
					}
					@{ @i @exp.0.iVal@ = (@exp.1.iVal@ / @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	"+" exp %prec UNARY	@{ @i @exp.0.iVal@ = + @exp.1.iVal@;
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	"-" exp %prec UNARY	@{ @i @exp.0.iVal@ = - @exp.1.iVal@;
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	exp "^" exp		@{ @i @exp.0.iVal@ = power (@exp.1.iVal@, @exp.2.iVal@);
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	|	"(" exp ")"		@{ @i @exp.0.iVal@ = @exp.1.iVal@;
					   @PV
					     {snprintf (node_label, sizeof(node_label), "exp|%d", @exp.0.iVal@);
					      agset (@exp.0.NIO@, "label", node_label);
					     }
					@}
	;

%%

#include <assert.h>

static int
power (int base, int exponent)
{
  int res = 1;
  assert (0 <= exponent);
  for (/* Niente */; exponent; --exponent)
    res *= base;
  return res;
}

#include "lexer.h"

result
parse (FILE *infile)
{
  yyscan_t scanner;
  yylex_init (&scanner);
  yyset_in (infile, scanner);
  result res = {1, 0, 0};
  yyparse (scanner, &res);
  yylex_destroy (scanner);
  return res;
}

#include <stdio.h>
#include <stdarg.h>

void
yyerror (yyscan_t scanner, result *res,
         const char *msg, ...)
{
  (void) scanner; /* avoid unreferenced parameter warning */
  va_list args;
  va_start (args, msg);
  vfprintf (stderr, msg, args);
  va_end (args);
  fputc ('\n', stderr);
  ++res->nerrs;
}

#include <getopt.h>

/* options descriptor */
static struct option opts[] = {
   { "input",   required_argument, NULL, 'i' },
   { "output",  required_argument, NULL, 'o' },
   { NULL,      0,                 NULL, 0   }
};

#include <libgen.h>
#include <stdlib.h>

int
main (int argc, char *argv[])
{
 char *inputFilename = (char *) NULL;
 char *DOTfilename = (char *) NULL;
 int ch;
 FILE *infile = stdin;

#if YYDEBUG
 /* enable parser runtime debugging */
 if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

 opterr = 0; /* disable getopt() generated error messages */
 while ((ch = getopt_long(argc, argv, ":i:o:", opts, NULL)) != -1) {
   switch (ch) {
     case 'i' : /* --input filename */
		inputFilename = optarg;
		break;
     case 'o' : /* --output filename */
		DOTfilename = optarg;
		break;
     case ':' : /* missing option argument */
		switch (optopt) {
		  case 'i' : /* missing --input filename  */
		  case 'o' : /* missing --output filename */
		  default  : /* something else is wrong   */
			     ;
		}
     case '?' : /* unknown option */
     default  : fprintf (stderr, "error in command line option: %s\n", argv[optind-1]);
		exit(EXIT_FAILURE);
   }
 }
 argc -= optind; argv += optind;
 /* process positional arguments */
 if (argc >= 1 && inputFilename == (char *) NULL)
    inputFilename = argv[argc-1];

 if (inputFilename != (char *) NULL)
   {
    infile = fopen (inputFilename, "r");
    if (!infile)
      {perror (inputFilename);
       exit(-1);
      }
   }

 result res = parse (infile);
 if (res.nerrs) return 1; /* failure if there were errors */

 if (inputFilename != (char *) NULL)
   {
    if (DOTfilename == NULL)
      {
       char *dir = strdup(dirname(inputFilename));
       char *base = strdup(basename(inputFilename));
       char *pt;

       strcat(dir, "/");
       pt = strrchr(base, '.');
       if (pt == (char *) NULL) /* append ".gv" to inputFilename */;
       else /* replace trailing extension with ".gv" */
	 *pt = '\0'; /* truncate inputFilename at the last '.' */
       DOTfilename = (char *)malloc(strlen(dir)+strlen(base)+1+2+1);
       strcat(strcat(strcpy(DOTfilename, dir), base), ".gv");
      }

    {FILE *out = fopen(DOTfilename, "w");
     agwrite(yyyTreeVizGraph, out);
     agclose(yyyTreeVizGraph); fclose(out);
    }
   }

 return 0;
}
