/* Infix notation calculator--calc */
%code requires /* -> header file, preceding the definition of YYSTYPE */
{
#include "types.h"
}

/* error count is exchanged between main, yyparse and yylex */
%param {int *nerrs}

%code provides /* -> header file, following the definition of YYSTYPE */
{
  /* the expected prototype of yylex:
   *   1st argument must be named `yylval' (because %define api.pure full)
   *   2nd argument must be named `nerrs'
   */
#define YY_DECL \
  enum yytokentype yylex (YYSTYPE *yylval, int *nerrs)
  YY_DECL;

  void yyerror (int *, const char *, ...);
}

%code
{
static int power (int, int);
}

%define api.pure full
%define api.push-pull both
%define api.token.prefix {TOK_}
%define api.value.type union
%define parse.error verbose
%define parse.trace
%verbose

%token
  EQ     "="
  PLUS   "+"
  MINUS  "-"
  STAR   "*"
  SLASH  "/"
  UP     "^"
  LPAREN "("
  RPAREN ")"
  EOL    "end-of-line"
  EOF 0  "end-of-file"
;

%token <int> NUM "number"
%type <int> exp
%printer { fprintf (yyo, "%d", $$); } <int>

/* precedence (lowest to highest) and associativity */
%nonassoc "="		/* comparison */
%left "+" "-"
%left "*" "/"
%precedence UNARY	/* unary plus or minus */
%right "^"		/* exponentiation */

@attributes {semantic_value iVal;} exp "number"
@attributes {@void print;} line

%%
input	:	line
	|	input line
	;

line	:	exp EOL			@{
					  @decorate
					  @e line.print: exp.iVal;
					     fprintf (stderr, "%d\n", @exp.iVal@);
					@}
	|	error EOL		{ yyyabort(); yyerrok; }
					@{ @m line.print; {} @}
//	|	EOL			@{ @m @line.print@ {} @} // TES: poor syntax errors (3) for this missing ';'
	|	EOL			@{ @m @line.print@; {} @}
	;

exp	:	"number"		@{ @i @exp.iVal@ = @"number".iVal@; @}
	|	exp "=" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ == @exp.2.iVal@); @}
	|	exp "+" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ + @exp.2.iVal@); @}
	|	exp "-" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ - @exp.2.iVal@); @}
	|	exp "*" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ * @exp.2.iVal@); @}
	|	exp "/" exp		{
					 if ($3 == 0)
					   {yyerror (nerrs, "invalid division by zero");
					    YYERROR;
					   }
					}
					@{ @i @exp.0.iVal@ = (@exp.1.iVal@ / @exp.2.iVal@); @}
	|	"+" exp %prec UNARY	@{ @i @exp.0.iVal@ = + @exp.1.iVal@; @}
	|	"-" exp %prec UNARY	@{ @i @exp.0.iVal@ = - @exp.1.iVal@; @}
	|	exp "^" exp		@{ @i @exp.0.iVal@ = power (@exp.1.iVal@, @exp.2.iVal@); @}
	|	"(" exp ")"		@{ @i @exp.0.iVal@ = @exp.1.iVal@; @}
	;

%%

#include <assert.h>

static int
power (int base, int exponent)
{
  int res = 1;
  assert (0 <= exponent);
  for (/* Niente */; exponent; --exponent)
    res *= base;
  return res;
}

#include "lexer.h"

#include <stdio.h>
#include <stdarg.h>

void
yyerror (int *nerrs, const char *msg, ...)
{
  va_list args;
  va_start (args, msg);
  vfprintf (stderr, msg, args);
  va_end (args);
  fputc ('\n', stderr);
  ++*nerrs;
}

#include <stdlib.h>

int
main (int argc, char *argv[])
{
 char *inputFilename = (char *) NULL;
 FILE *infile = stdin;
 int nerrs = 0;
 int status;
 yypstate *ps;

#if YYDEBUG
 /* enable parser runtime debugging */
 if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

 if (--argc >= 1) /* ignore the program name */
    inputFilename = argv[argc-1];

 if (inputFilename != (char *) NULL)
   {
    infile = fopen (inputFilename, "r");
    if (!infile)
      {perror (inputFilename);
       exit(-1);
      }
   }

 yyset_in (infile);

 ps = yypstate_new();
 do {
   YYSTYPE yylval;
   int yychar = yylex(&yylval, &nerrs);
   status = yypush_parse(ps, yychar, &yylval, &nerrs);
 } while (status == YYPUSH_MORE);
 yypstate_delete (ps);

 if (nerrs) return 1; /* failure if there were errors */

 return 0;
}
