/* $Id: symTrans.c 2098 2021-02-16 02:47:30Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <climits>
#include <cctype>
#else
#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#include <limits.h> 
#include <ctype.h>
#endif

#include "glob.h" 
#include "opts.h" 
#include "scan.h" 

extern FILE *yoxOutG;             /* the yacc-destined output */
extern FILE *yoxOutH;             /* the (optional) header file */ 
extern FILE *temp1Out; 


extern char *IDbuffer; 


extern char *gsBuffer;            /* lexeme of a grammar symbol */
extern int suffixBuffer;          /* holds the ? in: gramsym.?.attrib  */ 
extern char *atBuffer;            /* lexeme of an attribute name */
extern char *dumNodeSym; 
extern int stopAfterParse; 
 



/* Here follow some data structures for giving serial numbers to objects 
 * encountered in ox's input.  These sequences most of the time hold the 
 * value of the *next* serial number to be used, rather than the one that 
 * was *just* used. 
 */ 
/************************************************************************/ 
/* an increasing sequence of numbers to represent terminals of the grammar 
 */
#define firstTermNum 1 
long termSeq = firstTermNum; 



/* a decreasing sequence of numbers to represent nonterminals of the grammar 
 */
static int nontermSeq = -1; 



/* an increasing sequence of numbers to represent productions of the grammar 
*/
#define firstProdNum 1 
int prodNumSeq = firstProdNum; 


/* an increasing sequence of numbers to represent kinds of nodes of the 
   parse trees of the language 
 */  
#define NOTYPE 0 
#define firstStructTagNum (NOTYPE+1) 
static int structTagNumSeq = firstStructTagNum; 
/************************************************************************/


/* holds the identifier for the structure that Ox creates for each given 
   type of node  */ 
char **structTagArr;  



/* holds the identifier for the typedef for the above structure  */ 
char **structTDarr;  



/* Here follow some parameters to be set at evaluator generation time. 
 * Generally they may be changed from ox's command line. 
 */ 
/*******************************************************************/ 

/* The size of the array of pointers to "canonical" copies 
 * of lexemes (other than 
 * those for ox reserved words) found in ox's input.  
 * These strings will mostly be:
 * 1) identifiers and 2) strings that represent character constants.  
 */  
int maxNstrings = 509;  


/* The biggest permitted numbers of: grammar symbols, productions, 
 * attributes of one symbol, 
 * symbols in one production, kinds of parse tree nodes.  
 */  
int maxNgSyms = 509; 
int maxNprods = 1024; 
int maxNattrbsPerSym =  256;  
int maxProdLength = 256;  
int maxNstructTags = 256;
/*******************************************************************/ 


/* attributeIDlistBuffer[i] is a string that holds the name of 
 * the ith attribute of the kind of parse tree node currently 
 * being declared.  
 */  
char **attributeIDlistBuffer;  

/* points into attributeIDlistBuffer */   
int attrbIDlstBufPtr; 


/* stringTable[i] is either (char *)NULL or a canonical string 
 */ 
static char **stringTable; 


/* holds information about the current symbol in the current production
   until ready to install the symbol in the symbol table */  
struct gSym gSymBuffer; 


/* the table of grammar symbols */  
struct gSym *gSymTable;  



int bufPosInProd;  /* The position of an attribute occurrence in a rule */
int bufAttrbNum;   /* The serial number of the attribute in an occurrence */



/* points into prodBuffer2 */ 
int prodBufferPointer2 = 0; 

/* holds pointers to information about the symbols of the rule being 
   parsed
 */ 
struct pb2 *prodBuffer2; 

int errorProd; 

/*  generates and returns a pointer to a permanent null-terminated copy 
 *  of the first len characters of the string cp 
 */  
char *allocString(char *cp, int len) 
  {char *ncp, *str_save; 
   if ((ncp = ((char *) malloc(len+1))) 
       == 
       ((char *) NULL) 
      )  
      fatal("internal malloc error in allocString"); 
   str_save = ncp; 
   while (len--) 
         *ncp++ = *cp++; 
   *ncp = '\0';  
   return(str_save); 
  } 



/*  compute a hash value for a string */  
static int stringHash(char *cp, int len) 
  {int hashval; 
   switch (hashval = len) 
     {default: hashval += cp[len - 2] * 512; 
      case 3:  hashval += cp[len - 1] * 64; 
      case 2:  hashval += cp[1] * 8; 
      case 1:  hashval += *cp; break;  
      case 0:  fatal("internal error: can't seek the empty string.\n"); 
     }
   return(hashval); 
  } 



/* Search the string table for a string matching the first len characters of
 * cp.  If it is absent and insertion is requested, try to insert it, returning
 * a pointer to a permanent null-terminated copy if successful, null pointer 
 * if unsuccessful.  If absent and insertion not requested, return null 
 * pointer.  If present, return pointer to the null-terminated copy. 
 */  
char *stringLookup(char *cp, int len, int insertIfAbsent) 
  {int i,itry;  
   char **dum; 
   itry = stringHash(cp,len) % maxNstrings; 
   for (i=0; i<maxNstrings; i++) 
       {dum = stringTable + itry;  
        itry = (itry + 1) % maxNstrings;     
        if (*dum == (char *)NULL) 
           {if (insertIfAbsent) *dum = allocString(cp,len); 
            return(*dum); 
           } 
        if (strlen(*dum) != len) continue; /* don't match proper prefixes */ 
        if (!strncmp(*dum,cp,len)) return(*dum); 
       } 
   fatal("out of room in string table");  
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



/* for debugging */
void showStringTab(void) 
  {int i;  
   for (i=0; i<maxNstrings; i++) 
       if (stringTable[i] != ((char *)NULL))
          printf("%.5d   %s\n",i,stringTable[i]);
  } 



/* translate a representation of a character constant (read from ox's input) 
 * into the value of that constant on the machine on which ox is executing. 
 * This is useful when the target machine has the same character set as the 
 * machine on which ox is run.  For instance, if the C compilers on both 
 * machines use ASCII (for instance, when the ox-running machine
 * is the same machine as the target machine), then ox knows that 'A' 
 * means the same as '\201'.   
 */ 
unsigned char cctouc(char *cp) 
  {int temp; 
   if (strlen(cp) == 3) return((unsigned char) cp[1]); 
   switch (cp[2]) 
      {case 'n':  return((unsigned char) '\n'); 
       case 't':  return((unsigned char) '\t'); 
       case 'v':  return((unsigned char) '\v'); 
       case 'b':  return((unsigned char) '\b'); 
       case 'r':  return((unsigned char) '\r'); 
       case 'f':  return((unsigned char) '\f'); 
       case 'a':  yerror("unsupported escape sequence");
                  return((unsigned char) 'a'); 
       case '\\': return((unsigned char) '\\'); 
       case '?':  return((unsigned char) '?' ); 
       case '\'': return((unsigned char) '\''); 
       case '\"': return((unsigned char) '\"'); 
       default :  sscanf(cp,"\'\\%o\'",&temp);  
                  return((unsigned char) temp);  
      }   
  } 



/* stores the canonical textual representations of string representations
 * of character constants. Used exactly when CSCDMeqCSCM.  
 */ 
char *firstTextOfCC[UCHAR_MAX+1]; 
  


/* returns the first-encountered string representation of a character constant, 
   one of whose string representations has just been encountered. 
 */ 
char *getFirstTextOfCC(char *cp) 
  {unsigned char ind; 
   if (!CSCDMeqCSCM) return(cp);  
   ind = cctouc(cp); 
   if (firstTextOfCC[ind] == (char *)NULL) firstTextOfCC[ind] = cp; 
   return(firstTextOfCC[ind]);  
  } 



/* get the index into the grammar symbol table for a symbol, given 
 * its canonical string 
 */  
int gSymLookup(char *cp) 
  {int i,itry; 
   if (*cp == '\'') cp = getFirstTextOfCC(cp);  
   itry = (stringHash(cp,strlen(cp)) % maxNgSyms); 
   for (i=0; i<maxNgSyms; i++) 
       {if (gSymTable[itry].lexeme == cp) 
           return(itry); 
        if (gSymTable[itry].lexeme == (char *)NULL) return(-1); 
        itry = ((itry + 1) % maxNgSyms);  
       } 
   return(-1);   
  } 



/* This runs soon after each encounter of the @attributes reserved word. 
 * 
 *   {
 *    characterize this type of node by the current value of 
 *        structTagNumSeq after checking for overflow;
 *    get a block of characters to store the tag of this type of node, 
 *        such tag to be generated later in this routine, and such 
 *        tag to become part of a struct-specifier in the 
 *        generated C program.    
 *    get a block of characters to store a C identifier that will be 
 *        used in a typedef in the generated C program.  
 *    write the struct tag into the alotted space. 
 *    write the identifier for the typedef into its alotted space.  
 *    write the structure declaration to the Y-file.  
 *    write the structure declaration to the L-file.  
 *    get the node-type numbering sequence ready for next time.  
 *   } 
 */  
void initForAttrbDeclList(void) 
  {attrbIDlstBufPtr = 0; 
   if ((gSymBuffer.typeNum = structTagNumSeq) == maxNstructTags) 
      handleOverflow(MNST);   
   if ((structTagArr[structTagNumSeq] = (char *) malloc(10)) 
        == 
        ((char *) NULL) 
       )   
       fatal("malloc error in struct tag name allocation\n"); 
   if ((structTDarr[structTagNumSeq] = (char *) malloc(10)) 
        == 
        ((char *) NULL) 
       )   
       fatal("malloc error in struct tag name allocation\n"); 
   sprintf(structTagArr[structTagNumSeq],"yyyT%d",structTagNumSeq); 
   sprintf(structTDarr[structTagNumSeq],"yyyP%d",structTagNumSeq); 
   fprintf(yoxOutH,"\n\nstruct %s ",structTagArr[structTagNumSeq]);  
   structTagNumSeq++; 
  } 



/* inside ox, attributes declared for each type of node have attributes. 
 * These are called attribute attributes. This gets an array of structures,
 * one for each attribute of the current node type.  It then initializes, 
 * for each attribute of the current node, a certain one of its attributes
 * (viz. IorSknown). 
 * 
 */  
struct attrbAttrbs *allocAAlist(int len) 
  {int i; 
   struct attrbAttrbs *dum; 
   if ((dum = (struct attrbAttrbs *) malloc(len * sizeof(struct attrbAttrbs))) 
        != 
        ((struct attrbAttrbs *) NULL) 
       )   
       {for (i=0;i<len;i++) dum[i].IorSknown = 0; 
        return (dum); 
       } 
       else 
       fatal("malloc error in attribute attributes list allocation\n");  
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 


/* Things are complicated by permitting attribute declarations to 
   appear anywhere before the first %% mark.  There are different
   occasions for installing grammar symbols.  To simplify things,
   probably Ox should be changed such that attribute declarations 
   not appear before %token and %start.  
 */ 
int installGsymAfterAttrbDeclList(char *cp) 
  {int i,itry;
   gSymBuffer.lexeme = (cp[0] == '\'') ? getFirstTextOfCC(cp) : cp;  
   itry = (stringHash(gSymBuffer.lexeme,strlen(gSymBuffer.lexeme)) % maxNgSyms);
   for (i=0; i<maxNgSyms; i++) 
       {if ((gSymTable[itry].lexeme == gSymBuffer.lexeme) &&  
            (gSymTable[itry].typeNum != NOTYPE)
           )  
           {yerror("attributes redeclared for grammar symbol: "); 
            fprintf(stderr,"%s  %s.\n",gSymBuffer.lexeme,cp); 
            return(-1);  
           }  
        if ((gSymTable[itry].lexeme == gSymBuffer.lexeme) || 
            (gSymTable[itry].lexeme == (char *)NULL)
           )   
           {gSymTable[itry].lexeme = gSymBuffer.lexeme; 
            gSymTable[itry].attrbID = gSymBuffer.attrbID; 
            gSymTable[itry].nAttrbs = gSymBuffer.nAttrbs; 
            gSymTable[itry].typeNum = gSymBuffer.typeNum; 
            gSymTable[itry].aaList = allocAAlist(attrbIDlstBufPtr); 
            return(itry); 
           }  
        itry = ((itry + 1) % maxNgSyms);  
       } 
   fatal("full grammar symbol table\n"); 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



int installGsymAsTokenIfNew(char *cp) 
  {int i,itry;
   cp = (cp[0] == '\'') ? getFirstTextOfCC(cp) : cp;  
   itry = (stringHash(cp,strlen(cp)) % maxNgSyms);  
   for (i=0; i<maxNgSyms; i++) 
       {if (gSymTable[itry].lexeme == cp)   
           {if (gSymTable[itry].symNum == -1) 
               yerror("start symbol redeclared as token\n"); 
            if (gSymTable[itry].symNum == 0) 
               {if (termSeq == maxNterms) 
                   handleOverflow(NTRM);  
                gSymTable[itry].symNum = termSeq++; 
               }
            return(itry); 
           }  
           else 
           if (gSymTable[itry].lexeme == (char *)NULL) 
              {gSymTable[itry].lexeme = cp; 
               if (termSeq == maxNterms) 
                  handleOverflow(NTRM);  
               gSymTable[itry].symNum = termSeq++; 
               return(itry);  
              }  
        itry = ((itry + 1) % maxNgSyms);  
       } 
   fatal("full grammar symbol table\n"); 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



int installStartSym(char *cp) 
  {int i,itry;
   itry = (stringHash(cp,strlen(cp)) % maxNgSyms);
   for (i=0; i<maxNgSyms; i++) 
       {if (gSymTable[itry].lexeme == cp)   
           {if (gSymTable[itry].symNum > 0) 
               yerror("token redeclared as start symbol\n"); 
               else 
               gSymTable[itry].symNum = nontermSeq--;  
            return(itry); 
           }  
        if (gSymTable[itry].lexeme == (char *)NULL) 
           {gSymTable[itry].lexeme = cp; 
            gSymTable[itry].symNum = nontermSeq--; 
            return(itry);  
           }  
        itry = ((itry + 1) % maxNgSyms);  
       } 
   fatal("full grammar symbol table\n"); 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 
 


int installGsymAsNontermIfNew(char *cp) 
  {int i,itry;
   itry = (stringHash(cp,strlen(cp)) % maxNgSyms);
   for (i=0; i<maxNgSyms; i++) 
       {if (gSymTable[itry].lexeme == cp)   
           {if (gSymTable[itry].symNum != 0) 
               return(itry); 
               else  
               {gSymTable[itry].symNum = nontermSeq--; 
                return(itry);  
               }
           }  
        if (gSymTable[itry].lexeme == (char *)NULL) 
           {gSymTable[itry].lexeme = cp; 
            gSymTable[itry].symNum = nontermSeq--; 
            return(itry);  
           }  
        itry = ((itry + 1) % maxNgSyms);  
       } 
   fatal("full grammar symbol table\n"); 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 

 

void bufferAnAttrbID(char *cp) 
  {int i = 0; 
   if (attrbIDlstBufPtr == maxNattrbsPerSym) 
      fatal("too many attributes declared for a grammar symbol\n"); 
   if ((cp[0] == '\'') || isdigit(cp[0]))  
      yerror("syntax error in attribute identifier\n"); 
   while ((cp[i] != '.') && (cp[i] != '\0')) i++; 
   if (cp[i] != '\0') 
      yerror("syntax error in attribute identifier\n"); 
   attributeIDlistBuffer[attrbIDlstBufPtr++] = cp;  
  } 
 


/* Output the typedef for the node type just read. 
   Allocate a permanent list of attribute IDs 
   to be pointed to by the grammar symbol table entry.  Check
   whether, upon this node type, a new maximum number of attributes
   per node has been reached. Copy the buffer for the list of declared 
   attribute IDs into the permanent list, deleting duplicates.  
   Except in case of duplicates, original order of appearance in the 
   declaration list is preserved.
*/
void afterAttrbDeclList(void)  
  {char **dum; 
   int i,j; 
   fprintf(yoxOutH,"; \ntypedef struct %s *%s; \n",
                   structTagArr[gSymBuffer.typeNum], 
                   structTDarr[gSymBuffer.typeNum] 
          ); 
   if ((dum = (char **) malloc(attrbIDlstBufPtr * sizeof(char *))) 
       == 
       ((char **) NULL) 
      )   
      fatal("malloc error in attribute ID list allocation\n"); 
      if (attrbIDlstBufPtr > WAmax) WAmax = attrbIDlstBufPtr - 1; 
      for (i=0; i<attrbIDlstBufPtr; i++) 
        {for (j=0; j<i; j++)  
           if (dum[j] == attributeIDlistBuffer[i]) 
              {yerror(
                 "\n  attribute name used twice for same grammar symbol: "
                     );
               fprintf(stderr,"%s \n",dum[j]); 
               break;
              }  
         dum[i] = (j == i) ? attributeIDlistBuffer[i] : ((char *)NULL); 
        } 
      gSymBuffer.attrbID = dum; 
      gSymBuffer.nAttrbs = attrbIDlstBufPtr; 
  }  



#define maxNwyes 256 

void setupDumNodeSym(void) 
  {char wyes[maxNwyes]; 
   int i; 

   wyes[0] = 'y'; 
   for (i=1; i<maxNwyes; i++) 
       {wyes[i] = 'y'; 
        if (stringLookup(wyes,i,0) == (char *)NULL) break;  
       }
   if (i == maxNwyes) fatal("out of room for wyes\n");   
   dumNodeSym = stringLookup(wyes,i,1);  
   installGsymAsTokenIfNew(dumNodeSym);  
  }  



void printNodeSizeCalcTail(void);  
 


void afterAllAttrbDecls(void) 
  {int i;
   fprintf(temp1Out,"\n#define yyyCeiling(num,inc) ");   
   fprintf(temp1Out,"(((inc) * ((num)/(inc))) + (((num)%%(inc))?(inc):0)) \n"); 
   fprintf(temp1Out,"\n\n\nint yyyAlignSize = %d;",alignSize);   
   fprintf(temp1Out,"\nint yyyGNSz = sizeof(yyyGNT);");   
   fprintf(temp1Out,"\nint yyyNdSz[%d];\n",structTagNumSeq);   
   fprintf(temp1Out,"\nint yyyNdPrSz[%d];\n",structTagNumSeq);   
   fprintf(temp1Out,"\ntypedef int yyyCopyType;\n");   
   fprintf(temp1Out,"\nint yyyNdCopySz[%d];\n",structTagNumSeq);   
   fprintf(temp1Out,"long yyyBiggestNodeSize = 0;\n");  
   fprintf(temp1Out,"\nvoid yyyNodeSizeCalc()\n"); 
   fprintf(temp1Out,"  {int i;\n");  
   fprintf(temp1Out,"   yyyGNSz = yyyCeiling(yyyGNSz,yyyAlignSize); \n");   
   fprintf(temp1Out,"   yyyNdSz[0] = 0;\n");  
   for (i=1;i<structTagNumSeq;i++) 
       fprintf(temp1Out,"   yyyNdSz[%d] = sizeof(struct %s);\n",
                        i,structTagArr[i] 
              );  
   fprintf(temp1Out,"   for (i=0;i<%d;i++) \n",structTagNumSeq);   
   printNodeSizeCalcTail(); 
}



const char *nsText1[] = 
{  
"       {yyyNdSz[i] = yyyCeiling(yyyNdSz[i],yyyAlignSize); \n",
"        yyyNdPrSz[i] = yyyNdSz[i] + yyyGNSz;\n",
/*
"        fprintf(stderr,\"yyyNdPrSz[%d] == %d\\n\",i,yyyNdPrSz[i]);\n",   
*/ 
"        if (yyyBiggestNodeSize < yyyNdSz[i])\n",
"           yyyBiggestNodeSize = yyyNdSz[i];\n",
"        yyyNdCopySz[i] = yyyCeiling(yyyNdSz[i],sizeof(yyyCopyType)) / \n",
"                         sizeof(yyyCopyType); \n",
"       }\n",
"  }\n\n",
};

 

void printNodeSizeCalcTail(void) 
  {int i; 
   for (i=0;i < (sizeof(nsText1)/sizeof(char *));i++) 
     fprintf(temp1Out,"%s",nsText1[i]);
  } 



void incProdNumSeq(void) 
  {if ((++prodNumSeq) == maxNprods) 
      handleOverflow(NPRD);  
  } 

 

void putSymbolInProdBuffer2(int tabInd)             
  {if (prodBufferPointer2 > WSmax) WSmax = prodBufferPointer2;            
   if (prodBufferPointer2 < maxProdLength)            
      {prodBuffer2[prodBufferPointer2].lexeme = gSymTable[tabInd].lexeme; 
       prodBuffer2[prodBufferPointer2].gSymTabInd = tabInd; 
       if (!strcmp(gSymTable[tabInd].lexeme,"error")) errorProd = 1; 
       prodBufferPointer2++;
      } 
      else                                            
      handleOverflow(PRDL);   
  } 


/* for debugging the grammar-symbol table */
void showGsymInfo(int symPos) 
  {gSymType pSym;
   int j;  

   pSym = gSymTable + prodBuffer2[symPos].gSymTabInd;   
   fprintf(stderr,"     symbol Number: "); 
   fprintf(stderr,"%d\n",pSym->symNum); 
   fprintf(stderr,"     type :         "); 
   fprintf(stderr,"%d\n",pSym->typeNum); 
   fprintf(stderr,"     lexeme:        "); 
   fprintf(stderr,"%s\n",pSym->lexeme); 
   fprintf(stderr,"     number of attributes: ");
   fprintf(stderr,"%d\n",pSym->nAttrbs);  
   for (j=0;j<pSym->nAttrbs;j++) 
       {fprintf(stderr,"          attribute %2d: ",j); 
        fprintf(stderr,"%s",pSym->attrbID[j]);
        if (pSym->aaList[j].IorSknown) 
           {fprintf(stderr,"  line of first definition: "); 
            fprintf(stderr,"%d ",pSym->lineOfFirstDef); 
            if (pSym->aaList[j].InotS) 
               fprintf(stderr,"  inherited\n");  
               else 
               fprintf(stderr,"  synthesized\n");
           } 
           else 
           fprintf(stderr,"error: neither syn nor inh \n");  
       } 
  } 



void showSymInfoForProd(void) 
  {int k; 

   fprintf(stderr,"\n\n\nSymbol information for production %d : \n",
           prodNumSeq - 1
          ); 
   for (k=0;k<prodBufferPointer2;k++) 
       {fprintf(stderr,"Position: %d\n",k); 
        showGsymInfo(k);
       } 
  } 



/* translates (gsBuffer,suffixBuffer,atBuffer) into 
              (bufPosInProd,bufAttrbNum) 
*/
void transAttrbNameToNumPair(void) 
  {struct gSym dum; 
   int nTimesSymSeen = 0; 

   bufPosInProd = 0; 
   bufAttrbNum = 0; 
   if (suffixBuffer == -1) goto err2;  
   while (bufPosInProd<prodBufferPointer2) 
     {if (prodBuffer2[bufPosInProd].lexeme == gsBuffer) 
         {if (suffixBuffer == nTimesSymSeen) 
             break; 
             else 
             nTimesSymSeen++;
         }  
      bufPosInProd++; 
     }    
   if (bufPosInProd == prodBufferPointer2) goto err; 
   dum = gSymTable[prodBuffer2[bufPosInProd].gSymTabInd];  
   while (bufAttrbNum < dum.nAttrbs) 
     {if (dum.attrbID[bufAttrbNum] == atBuffer) 
         return;
         else
         bufAttrbNum++;
     }
err:
   yerror("attribute name not valid for this production: \n    "); 
   fprintf(stderr,"%s.%d.%s\n",gsBuffer,suffixBuffer,atBuffer);  
err2: 
   bufPosInProd = -1; 
   bufAttrbNum = -1; 
   return;  
  } 



void prepareDataStructures1(void) 
  {int i;

   if ((stringTable = (char **) malloc(maxNstrings * sizeof(char *)))  
       != 
       ((char **) NULL) 
      )   
      for (i=0; i<maxNstrings; i++) stringTable[i] = (char *)NULL; 
      else 
      fatal("malloc error in string table allocation\n");  

   if ((gSymTable = 
        ((struct gSym *) malloc(sizeof(struct gSym) * maxNgSyms )) 
       )  
       != 
       ((struct gSym *) NULL) 
      )   
      for (i=0; i<maxNgSyms; i++) 
          {gSymTable[i].symNum = 0; 
           gSymTable[i].lexeme = (char *)NULL; 
           gSymTable[i].nAttrbs = 0; 
           gSymTable[i].typeNum = NOTYPE; 
           gSymTable[i].lineOfFirstDef = -1; 
          } 
      else 
      fatal("out of memory for grammar symbol table \n"); 

   for (i=0;i<=UCHAR_MAX;i++) firstTextOfCC[i] = (char *)NULL;  
  
   if ((structTagArr = (char **) malloc(maxNstructTags * sizeof(char *)))  
       == 
       ((char **) NULL) 
      )   
      fatal("malloc error in struct tag string array allocation\n");  

   if ((structTDarr = (char **) malloc(maxNstructTags * sizeof(char *)))  
       == 
       ((char **) NULL) 
      )   
      fatal("malloc error in struct typedef string array allocation\n");  

   if ((attributeIDlistBuffer = 
        (char **) malloc(maxNattrbsPerSym * sizeof(char *))
       )  
       == 
       ((char **) NULL) 
      )   
      fatal("malloc error in attribute ID list buffer allocation\n");  
   
   if ((prodBuffer2 = 
        (struct pb2 *) malloc(maxProdLength * sizeof(struct pb2))
       )  
       == 
       ((struct pb2 *) NULL) 
      )   
      fatal("malloc error in production buffer allocation\n");  
   
  }


